#! /usr/local/bin/perl -w
#                           @(#)  %version: 1.28 % %derived_by: guidod %
#                           %date_modified: Mon Jan 31 16:35:57 2000 %

#  # a comment for doc-pfe will look like this:
#  /** FORTH-NAME ( in-var1 in-val2 -- out-val ) some small notes
#    * some text that will all go
#    * into one paragraph, so this
#    * text is probably on one line
#    example: code is written without a start-prefix
#      this text here is then surrounded by <pre>
#      so it is taken (almost) verbatim
#    * but you can also switch back and include a reference
#    * to a forth-name using => ANOTHER-FORTH-NAME
#    */                 # the end-of-comment must be alone on the line;
# FCode(p4_forth_name)  # the doc-pfe knows about FCode and can
# {                     # associate the above comment (and name) with it
# ....
# }
#                         # the doc-pfe knows about LISTWORDS
# LISTWORDS(forth_extra)  # it will scan the items for cross reference
# {                       # and create a file wordset-forth_extra
#   /** in-place comment for all the directly following declarations */
#   CO("FORTH-NAME", p4_forth_name),   # for each of these wordsets
# };                                   # the header will contain this...
# COUNTWORDS(forth_extra, "FORTH-EXTRA - the true identity");
#
# you can also add extra entries that do look like
# /** some comment like the above
#  */
# /*"an-export-name"*/
# the export-name can then be resolved everywhere, and the definitions
# will go into an extra series of documentation, since these docs are
# thought to be neither representing C nor Forth Code. 

##########################################################
# declare the variables and databases used

my $version = "doc-pfe 1.0";
my $helpstring = "\n"
  ."--dpans=dirpath make crossreference to ANSI Forth Directory \n" 
  ."--each=prefix   create a doc-file for each exported name \n"
  ."                (default prefix is 'w-')\n"
  ."--each-index=file a list of filenames generated for --each\n"
  ."--src-index     generate a file per source\n"
  ."--src-index=file put the list of src-symbols here (index-symbols.htm)\n"
  ."--help          this help \n"
  ."--version       the version \n"
  ."--warn-nodoc    warn when no explicit documentation available \n"
  ."                     for a generated symbol \n"
  ."--warn-extra    warn when an extra export documentation available \n"
  ."                     for a generated symbol already in a wordset \n"
  ."--debug         do some debugging \n"
  ."--debug-*       debug for a specific internal table \n"
  ."--refs          with references to the source-files";

my @files;

# the database is built using perl-hashes - sometimes a trick
# is used to make the keys unique. Even more I have no knowledge
# about how to put structures as the value of a hash - would be
# a nice thing to replace wordset_a_* with a neater thing...
my %ccode; # ccode function -> filename:$lineno of its definition
my %fcode; # fcode function -> filename:$lineno of its definition
my %ccode_comment; # ccode function -> comment text
my %fcode_comment; # fcode function -> comment text
my %wordset; # wordset -> filename of its definition
my %wordset_name; # wordset -> description string
my %word_export; # forth-visible name -> array of wordset:lineno 
my %wordset_a_kind; # wordset:lineno -> a1  ie. the forth-export kind
my %wordset_a_export; # wordset:lineno -> a2  ie. the forth-visible name
my %wordset_a_symbol; # wordset:lineno -> the c-visible symbol
my %qcode; # export-name -> filename:$lineno of its definition
my %qcode_comment; # export-name ->  comment text

my %section_file; # sectionname -> filename
my %section_comment; # sectionname -> complete comment

##########################################################
# seperate the files from other options on the commandline

my %opt = ();

foreach (@ARGV) {
  if  (/^--\w.*/) {
  	do { print "$version\n"; exit 0; } if /^--version$/;
  	do { print "$version help\n$helpstring\n"; exit 0; } if /^--help$/;
        if (/^--([-\w]+)=(.*)$/)
        {
          $opt{$1} = $2;
        }else{
          s/^--//;
          $opt{$_} = "";
        }
  }else{
  	push @files, $_;
  }
}

my $debug = defined $opt{"debug"} ? 1 : 0;

##########################################################
# scan the files and build the database(s)

foreach $F (@files)
{
  my $state = "normal";
  my $lineno = "0000";
  open F, $F or do { print STDERR "could not open $F\n"; next; };

  while (<F>)
  {
    $lineno++;
    if ($state eq "normal") {
    	if (m/ ^\s* \/\*\" (.*) \"\*\/ \s*$ /x) {
		$qcode{$1} = "$F:$lineno"; 
		if (defined $comment) {
			$qcode_comment{$1} = $comment;
			# does *not* undef comment
		}
		next;
	}
        if (m/ ^\s* \/\*+ \s* \@\{ \s* \*\/ \s*$ /x) {
		if (defined $comment and $comment =~ m/\@name(.*)$/m) {
	                if (length $1) { my $secname = $1;
				$secname =~ s/^\s*//; 
     				$secname =~ s/\s*$//;
				$section_file{$secname} = $F;
				$section_comment{$secname} = $comment;
			}
			undef $comment;
		}
		next;
        }
    	if (m/ ^\s* FCode\s*\(\s* (\w+) \s*\) /x) {
		$fcode{$1} = "$F:$lineno"; 
		if (defined $comment) {
			$fcode_comment{$1} = $comment;
			undef $comment;
		}
		next;
	}
    	if (m/ ^\s* (p4_\w+) \s*\( /x) {
		$ccode{$1} = "$F:$lineno"; 
		if (defined $comment) {
			$ccode_comment{$1} = $comment;
			undef $comment;
		}
		next;
	}
        if (m/ ^\s* \/\*\* \s* (.*) \s* \*\/ \s* $ /x) {
		$comment = "*\n* ".$1;
		next;
	}
	if (m/ ^\s* \/\*\* \s+ (.*) /x) {
		undef $comment;
		$comment = "* $1";
		$state = "comment";
		next;
  	}
	if (m/ ^\s* (P4_)?LISTWORDS \s*\(\s* (\w+) \s*\)\s*\= /x) {
		$wordset{$2} = $F;
		$this_wordset = "$2";
		$state = "wordset";
		undef $comment;
		next;
  	}
	if (m/ ^\s* (P4_)?COUNTWORDS \s*\(\s* (\w+) \s*,\s* \"([^\"]+)\" \s*\) /x) {
		$wordset_name{$2} = $3;
		next;
  	}
    }
    elsif ($state eq "comment")
    {
	if (m/ ^\s* \*\/ \s* \/\*\" (.*) \"\*\/ /x) {
			$qcode{$1} = "$F:$lineno";
			$qcode_comment{$1} = $comment if defined $comment; 
			# does *not* undef $comment
			$state = "normal"; next;
	}
	if (m/ ^\s* \*\/ /x) { $state = "normal"; next;	}
	if (m/ ^\* (.*) /x) { $comment .= "\n* $1"; next; }
	if (m/ ^\s\* (.*) /x) { $comment .= "\n* $1"; next; }
	if (m/ ^\s\s\* (.*) /x) { $comment .= "\n* $1"; next; }
        chop;
	$comment .= "\n$_";
    }
    elsif ($state eq "wordset")
    {
        if (m/ ^\s* (\w\w) \s*\(\s* \"(.*)\" \s*,\s* (\w[^(),\ ]+) \s*\)\s*, /x) {
		push @{$word_export{$2}}, "$this_wordset:$lineno";
		$wordset_a_kind{"$this_wordset:$lineno"} = $1;
		$wordset_a_export{"$this_wordset:$lineno"} = $2;
		$wordset_a_symbol{"$this_wordset:$lineno"} = $3;
		
		$wordset_a_export{"$this_wordset:$lineno"} =~ s(\\([^\\]))($1)g;
		if (defined $comment and not defined $fcode_comment{$3}) {
			$fcode_comment{$3} = $comment;
		}
		next;
	}
        if (m/ ^\s* (\w\w) \s*\(\s* \"(.*)\" \s*,\s* ([^(),\ ]+) \s*\)\s*, /x) {
		push @{$word_export{$2}}, "$this_wordset:$lineno";
		$wordset_a_kind{"$this_wordset:$lineno"} = $1;
		$wordset_a_export{"$this_wordset:$lineno"} = $2;
		$wordset_a_symbol{"$this_wordset:$lineno"} = "$this_wordset.$lineno";
		
		$wordset_a_export{"$this_wordset:$lineno"} =~ s(\\([^\\]))($1)g;
		if (defined $comment) {
			$fcode_comment{"$this_wordset.$lineno"} = $comment;
		}
		next;
	}
        if (m/ ^\s* (\w\w) \s*\(\s* \"(.*)\" \s*\)\s*, /x) {
		push @{$word_export{$2}}, "$this_wordset:$lineno";
		$wordset_a_kind{"$this_wordset:$lineno"} = $1;
		$wordset_a_export{"$this_wordset:$lineno"} = $2;
		$wordset_a_symbol{"$this_wordset:$lineno"} = "$this_wordset.$lineno";
		
		$wordset_a_export{"$this_wordset:$lineno"} =~ s(\\([^\\]))($1)g;
		if (defined $comment) {
			$fcode_comment{"$this_wordset.$lineno"} = $comment;
		}
		next;
	}
        if (m/ ^\s* \/\*\* \s* (.*) \s* \*\/ \s* $ /x) {
		$comment = "*\n* ".$1;
		next;
	}
	undef $comment;
	if (m/ ^\s* \}\; /x) {
		$state = "normal";
		next;
	}
    }
  }
  close F;
}

sub canonical_forth_symbol
{
    $_ = $_[0];
    while (s/_*\(([^\(\)]*)\)_*/_paren_$1_/g) {}
    while (s/_*\[([^\[\]]*)\]_*/_bracket_$1_/g) {}
    while (s/_*\<([^\<\>]*)\>_*/_local_$1_/g) {}
    s/_*\?_*/_Q_/g; s/_*\:_*/_colon_/g; s/_*\;_*/_semicolon_/g;
    s/_*\#_*/_sh_/g; s/_*\+_*/_plus_/g; s/_*\,_*/_comma_/g;
    s/_*\!_*/_store_/g; s/_*\@_*/_fetch_/g; s/_*\"_*/_quote_/g;
    s/_*\<_*/_less_/g; s/_*\=_*/_equal_/g; s/_*\>_*/_to_/g;
    s/_*\/_*/_slash_/g; s/_*\*_*/_star_/g; s/_*\._*/_dot_/g;
    s/_*\'_*/_tick_/g; s/_*\-_*/_minus_/g; s/_*\~_*/_not_/g;
    s/_*\&_*/_and_/g; s/_*\|_*/_or_/g; s/_*\^_*/_xor_/g;
    s/_*\\_*/_backslash_/g; s/_*\`_*/_backtick_/g; 
    s/_*\{_*/_bbegin_/g; s/_*\}_*/_endb_/g; 
    s/_*\(_*/_pbegin_/g; s/_*\)_*/_endp_/g; 
    s/_*\[_*/_mbegin_/g; s/_*\)_*/_endm_/g; 
    s/__/_/g;
    return $_;
}

# ---------------- create "extra" wordset --------------------
# build a virtual "extra" wordset from the qcode entries
{ # we need to make up symbols that are valid enough to
  # be used as html section reference names, and possibly filenames.
  # so in fact, they must be just like filenames. To ease things up,
  # we're replacing everything not intelligible with a ~ (tilde) and
  # then replace any trailing and preceding sequences of tildes. 
  # A panicking message will be shown, if that will lead us into 
  # trouble with other symbols already there (although we continue).
  my $extra = "noexport";
  my $virtual_lineno = 1;
  qcode:
  for $w (keys %qcode)
  {
    $_ = canonical_forth_symbol($w);
    y/A-Za-z0-9\_\-/\~/c;
    s/^~/_/;
    s/~$/_/;     
    if (not length) { 
	print STDERR "PANIC: symbol gone empty (while generating symbol from \"$w\") - skipped\n";
	next;
    }
    if (exists $fcode{$_}) {
	print STDERR "PANIC: symbol $_ exists (while generating symbol from \"$w\") - skipped\n";
        next;
    }else{
        my $we = $w; $we =~ s(\\([^\\]))($1)g;
        for my $k (keys %wordset_a_export)
        {
	   if ($wordset_a_export{$k} eq $we) {
	      if (not defined $fcode_comment{$wordset_a_symbol{$k}}) {
		 $fcode{$wordset_a_symbol{$k}} = $qcode{$w};
		 $fcode_comment{$wordset_a_symbol{$k}} = $qcode_comment{$w};
		 print STDERR "$w already exported in $k - added comment \n"
		    if exists $opt{"warn-extra"};
	      }else{
		 print STDERR "$w already exported in $k - skipped \n"
		    if exists $opt{"warn-extra"};
	      }
	      next qcode;
	   }
        }

       	$fcode{$_} = $qcode{$w};
       	$fcode_comment{$_} = $qcode_comment{$w};
       	push @{$word_export{$we}}, " $extra:$virtual_lineno";
       	$wordset_a_kind{"$extra:$virtual_lineno"} = "never";
       	$wordset_a_export{"$extra:$virtual_lineno"} = $we;
       	$wordset_a_symbol{"$extra:$virtual_lineno"} = $_;
    }
    $virtual_lineno++;
  } 
  print STDERR "$virtual_lineno extra definitions added \n";
  $wordset{"$extra"} = "$extra.definitions";
  $wordset_name{"$extra"} = "~ extra definitions which may be not exported";
}

####################################################################
# create cross-references - so make even more database(s) entries

my %export_name; # export_name -> wordset:$lineno
my %export_name_file; # export_name -> filenamebase
my %export_name_from_file; # filenamebase -> export_name

foreach my $W (keys %wordset_a_export)
{
  ## first part ##
  my $name = $wordset_a_export{$W};
  $export_name{$name} = $W;	

  ## second part ##
  {  my $prefix = "_";
     $prefix = $opt{"each"} if exists $opt{"each"} && length $opt{"each"};
     my $f = $prefix.$wordset_a_symbol{$W}."-".lc($wordset_a_kind{$W});

     $f =~ tr/0-9A-Za-z\_\w\-/\-/cs;

        if (!exists $export_name_from_file{$f})
        {
           $export_name_file{$name} = $f;
           $export_name_from_file{$f} = $name;
        }else{
            my $i = 2;
            while (exists $export_name_from_file{$f."-".$i}) { ++$i; }
            $export_name_file{$name} = $f."-".$i;
            $export_name_from_file{$f."-".$i} = $name;
        }
  }
}				

## print "\n\n export_names:\n", join ("\n",%export_name); # DEBUG

sub crossref_export_name
{
  ## print STDERR "find for $_[0]\n";
  if (exists $export_name{$_[0]}) {
	my $ws = $export_name{$_[0]};
        my $w = $wordset_a_symbol{$ws} if exists $wordset_a_symbol{$ws};
        $ws =~ s/:.*//; # chop linenumber
        if (defined $w) {
        	return "wordset-$ws.html#$w";
	}else{
		return "wordset-$ws.html";
	}
  }else{
	return "index-words.html";
  }
}

sub replace_crossreferences
{
  $_ = $_[0]; 

  s(\=\>\s([^\s]+))("<a href=\"".&crossref_export_name($1)."\"><c> ".&v($1)."</c></a>")ge;
  s(\=\>'([^']+)')("<a href=\"".&crossref_export_name($1)."\"><c>".&v($1)."</c></a>")ge;

  return $_;
}

my %nonref_export_name_file;
sub crossref_export_name_file
{
   $_ = $_[0];
   my $def = $export_name_file{$_};
   return $def if defined $def;
   if (not exists $nonref_export_name_file{$_})
   { print STDERR "# no crossref for $_ (using index-each.htm now)\n";
     $nonref_export_name_file{$_} = 1;
   }
   return "index-each";  
}
   

sub replace_crossreferences2
{
  $_ = $_[0]; 

  s(\=\>\s([^\s]+))("<a href=\"".&crossref_export_name_file($1).".html\"><c> ".&v($1)."</c></a>")ge;
  s(\=\>'([^']+)')("<a href=\"".&crossref_export_name_file($1).".html\"><c>".&v($1)."</c></a>")ge;

  return $_;
}

# ------------------------- wordset_shortname ----------------------
my %wordset_shortname ; # wordset -> official wordset name as with ENVIRONMENT

for my $W (keys %wordset_name)
{
    $_ = $wordset_name{$W};
    s/([^ ]*) .*/$1/ if (/ /);
    s/-EXT$// if (/-EXT$/);
    $_ = "" if length $_ < 2 ;
    $wordset_shortname{$W} = $_;
}

#####################################################################
# check for the ansi standard files

my %dpans;

if (defined $opt{"dpans"} && length $opt{"dpans"})
{
  my $F = $opt{"dpans"}."/dpansf.html";
  if (not open F, "$F") 
  { 
	print STDERR "cannot open dpans index file $F - skipped dpans\n";
  }else{
      print STDERR "reading dpans...";
      while (<F>)
      {
        if (m(^\s+ 
             (\<A \s+ href=dpans\d+\.html\#[\d\.]+\> [\d\.]+ \<\/A\>) 
             \s+ ([^\s]+) \s)ix
        )
        {
          my $ref = $1;
          my $key = $2;
     
          $key =~ s(<B>.*</B>)();
          $key =~ s(&lt;)(<)g;
          $key =~ s(&gt;)(>)g;
      
          $ref =~ s/href=dpans/"href=".$opt{"dpans"}."\/dpans"/e;
    
          $dpans{$key} = $ref;
        }
      }

      print STDERR "got ",scalar(%dpans)," definitions\n";
      close F;
  }
}

#####################################################################
# dump the databases for debugging purpose

if ($debug or exists $opt{"debug-dpans"}) {
  my $F = "doc-pfe.dpans";
  open F, ">$F";
  print STDERR "created $F\n";
  for (sort keys %dpans) {
  	print F "\n",$_,"  ",$dpans{$_}; 
  }
  close F;
}
if ($debug or exists $opt{"debug-ccode"}) {
  my $F = "doc-pfe.ccode";
  open F, ">$F";
  print STDERR "created $F\n";
  for (sort keys %ccode) {
  	print F "\n",$_,"  ",$ccode{$_}; 
  }
  close F;
}
if ($debug or exists $opt{"debug-fcode"}) {
  my $F = "doc-pfe.fcode";
  open F, ">$F";
  print STDERR "created $F\n";
  for (sort keys %fcode) {
  	print F "\n",$_,"  ",$fcode{$_}; 
  }
  close F;
}
if ($debug or exists $opt{"debug-ccode-comment"}) {
  my $F = "doc-pfe.ccode-comment";
  open F, ">$F";
  print STDERR "created $F\n";
  for (sort keys %ccode_comment) {
  	print F "\n",$_,"  ",$ccode_comment{$_}; 
  }
  close F;
}
if ($debug or exists $opt{"debug-fcode-comment"}) {
  my $F = "doc-pfe.fcode-comment";
  open F, ">$F";
  print STDERR "created $F\n";
  for (sort keys %fcode_comment) {
  	print F "\n",$_,"  ",$fcode_comment{$_}; 
  }
  close F;
}
if ($debug or exists $opt{"debug-wordsets"}) {
  my $F = "doc-pfe.wordset";
  open F, ">$F";
  print STDERR "created $F\n";
  for (sort keys %wordset_name) {
  	print F "\n",$_,"  ",$wordset_name{$_}; 
  }
  close F;
}
if ($debug or exists $opt{"debug-exportname"}) {
  my $F = "doc-pfe.exportname";
  open F, ">$F";
  print STDERR "created $F\n";
  for (sort keys %export_name) {
  	print F "\n",$_,"  ",$export_name{$_}; 
  }
  close F;
}
 
######################################################################
# preformat and internal generation of text

$style_nounder = "<style> a:link { text-decoration:none } </style>";

# we're given a forth-declaration - ie. the first line in that comment
# FORTH-NAME ( in1 in2 -- out ) some more comment
sub format_fcode_header
{
  my $a_name = "";
  my $a_stack = "";
  my $a_comment = "";
  if ($_[0] =~ m/ ^(.*) (\(.*--.*\)) (.*) /x) {
        $a_comment = $3;
        $a_stack = $2;
        $a_name = $1; 
  }elsif
     ($_[0] =~ m/ ^(.*) (\([^()]*\)) (.*) /x) {
        $a_comment = $3;
        $a_stack = $2;
        $a_name = $1; 
  }else{
        $a_name = $_[0];
  }
  $a_name =~ s/^\s+//; 
  $a_name =~ s/\s+$//;
  $a_name = $_[1] if ! length $a_name;

  return "<code><b> ".v($a_name)." </b> ".v($a_stack)." </b></code> &nbsp; <i>".$a_comment."</i>";
}

sub kind_name 
{
  $_ = $_[0];
  /CO/ && return "";
  /CS/ && return "smart-word";
  /CI/ && return "immediate";
  /CX/ && return "on-load";
  /CY/ && return "on-exit";
  /OV/ && return "variable";
  /OC/ && return "constant";
  /OL/ && return "value";
  /VO/ && return "vocabulary";
  /IV/ && return "immediate-variable";
  /IC/ && return "immediate-constant";
  /IL/ && return "immediate-value";
  /SV/ && return "system-variable";
  /SC/ && return "system-constant";
  /DV/ && return "dict-variable";
  /DC/ && return "dict-constant";
  return "...";
}

sub v # print verbatim w.r.t HTML syntax
{
  $_ = $_[0];

  s/\&/\&amp;/g;
  s/\</\&lt;/g;
  s/\>/\&gt;/g;
  s/\"/\&quot;/g;

  return $_;
}

sub o # replace non-html tags with their output representation
      # this may be the place to put other output formats in
{
  $_ = $_[0];
  s:<c>:<nobr><code>:sg;        s:</c>:</code></nobr>:sg;
  s:<bc>:<nobr><b><code>:sg;    s:</bc>:</code></b></nobr>:sg;

  # common html-errors in user-given text
  s:\<=:\&lt;=:g ;
  s:\< :\&lt; :g ;
  s:\& :\&amp; :g ;
  return $_;
}

# ................... pre-format fcode entries ......................

my %fcode_head = (); # just the line, no &format_fcode_header
my %fcode_body = (); # simple format, no &replace_crossreferences

for my $symbol (keys %fcode_comment)
{
  my $comment = $fcode_comment{$symbol};
  my $s = "firstline"; # the state - for the following state-machine...
  my $head = "";
  my $body = "";

    {
        foreach my $line (split("\n", $comment)) 
	{
            print LOG "\n+++: $line" if $debug;
	    if ($s eq "firstline") {
		    $line =~ s/^\*//;
                    $head .= $line;
		    $s = "normal";
		    next;
	    }
	    if ($s eq "normal") {
		    if ($line !~ m/^\*/) {
			    $body .= "<pre>".v($line)."\n";
			    $s = "pre";
		    }else{
			    $line =~ s/^\*//;
			    $line = "<p>" if $line =~ m/^\s*$/;
			    $body .= $line ."\n";
		    }
		    next;
	    }
	    if ($s eq "pre") {
		    if ($line !~ m/^\*/) {
			    $body .= v($line)."\n";
		    }else{
			    $body .= "</pre>\n";
			    $line =~ s/^\*//;
			    $body .= $line ."\n";
			    $s = "normal";
		    }
		    next;
	    }
        }
    	if ($s eq "pre") { $body .= "</pre>\n"; }
	$fcode_head{$symbol} = $head;
        $fcode_body{$symbol} = $body;
    }
}

# ................... pre-format ccode entries ......................

my %ccode_head = (); # just the line, no "$symbol :" before, no v()
my %ccode_body = (); # simple format, no &replace_crossreferences

for my $symbol (keys %ccode)
{
  my $comment = $ccode_comment{$symbol};
  my $s = "firstline"; # the state - for the following state-machine...
  my $head = "";
  my $body = "";

    if (defined $comment) {
        foreach my $line (split("\n", $comment)) 
	{
            print LOG "\n+++: $line" if $debug;
	    if ($s eq "firstline") {
		    $line =~ s/^\*//;
                    $head .= $line;
		    $s = "normal";
		    next;
	    }
	    if ($s eq "normal") {
		    if ($line !~ m/^\*/) {
			    $body .= "<pre>".v($line)."\n";
			    $s = "pre";
		    }else{
			    $line =~ s/^\*//;
			    $line = "<p>" if $line =~ m/^\s*$/;
			    $body .= $line ."\n";
		    }
		    next;
	    }
	    if ($s eq "pre") {
		    if ($line !~ m/^\*/) {
			    $body .= v($line)."\n";
		    }else{
			    $body .= "</pre>\n";
			    $line =~ s/^\*//;
			    $body .= $line ."\n";
			    $s = "normal";
		    }
		    next;
	    }
        }
    	if ($s eq "pre") { $body .= "</pre>\n"; }
	$ccode_head{$symbol} = $head;
        $ccode_body{$symbol} = $body;
    }
}

# ....................... preformat section desc ......................
my %section_header; # sectionname -> description
my %section_nohead; # sectionname -> other infos (plain)
my %section_from_shortname; # wordsetlike-shortname -> sectionname

for my $section_name (keys %section_comment)
{
  my $comment = $section_comment{$section_name};
  my $s = "firstline"; # the state - for the following state-machine...
  my $header = "";
  my $nohead = "";

    {
        foreach my $line (split("\n", $comment)) 
	{
            print LOG "\n+++: $line" if $debug;
	    if ($s eq "firstline") {
		    # don't take the first line 
                    # and go for "extra state"
		    $s = "extra";
		    next;
	    }
            if ($s eq "extra") {
		if ($line =~ m/\@description/) {
			$s = "normal";
		}else{
			$nohead .= $line ."\n";
		}
		next;
	    }
	    if ($s eq "normal") {
		    if ($line =~ m/^\s*\*\s*\@see/
                    or  $line =~ m/^\s*\s\s*\@author/
                    or  $line =~ m/^\s*\s\s*\@version/ ) {
			$nohead .= $line ."\n";
			$s = "extra";
			next;
		    }


		    if ($line !~ m/^\*/) {
			    $header .= "<pre>".v($line)."\n";
			    $s = "pre";
		    }else{
			    $line =~ s/^\*//;
			    $line = "<p>" if $line =~ m/^\s*$/;
			    $header .= $line ."\n";
		    }
		    next;
	    }
	    if ($s eq "pre") {
		    if ($line !~ m/^\*/) {
			    $header .= v($line)."\n";
		    }else{
			    $header .= "</pre>\n";
			    $line =~ s/^\*//;
			    $header .= $line ."\n";
			    $s = "normal";
		    }
		    next;
	    }
        }
    	if ($s eq "pre") { $header .= "</pre>\n"; }
	$nohead =~ s/\@\(\#\)//; $nohead =~ s/\%(\w)/\% $1/g; # continuus
	$section_header{$section_name} = $header;
        $section_nohead{$section_name} = $nohead;

	$_ = $section_name;
        m/^\s*([^ ]+) .*$/;
	$section_from_shortname{$1} = $section_name if defined $1;
    }
}

for (sort keys %section_from_shortname)
{	
	print STDERR "($_ : $section_from_shortname{$_})";
}


#####################################################################
# now start generating files

# .................... index-words.html .............................

for my $F ("index-words.htm")
{
   open F, ">$F" or next;
   print STDERR "created $F\n";

   print F "<html><head> <meta name=\"Author\" content=\"$ENV{USER}\">\n";
   print F "<meta name=\"date\" content=\"",scalar localtime(),"\">\n";
   print F "\n<title> Exported Names in Wordsets </title>\n";
   print F "\n", $style_nounder;
   print F "\n</head><body>";
   print F "\n";
   print F "\n<h2> Exported Names </h2>";
   print F "\n";
   print F "<table border=0 cellspacing=0 cellpadding=0>\n";
   
   for my $W (sort keys %export_name)
   { 
        $_ = $export_name{$W}; s/:.*//; 
        my $ws_shortname = $wordset_shortname{$_}; 

	print F "<tr valign=top><td><small><i> ",
                " ", $ws_shortname, " &nbsp;&nbsp;</i></small></td>";
        print F "<td><a href=\"", &crossref_export_name($W),"\"><code>", 
                " ", v($W), " </code></a></td></tr>\n",
   }
   print F "</table>\n";
   print F "\n</body></html>\n";
   close F;	
}

# .....................  wordset-*.html  ...........................

my %export_name_each; # does now also create %export_name_each documents

foreach my $W (keys %wordset)
{
  my ($ws_name, $ws_shortname);
  $F = "wordset-$W.htm";
  $LOG = "wordset-$W.log";
  # if you add more intermediate text-storages, keep the names
  # in alphabetical order with their spit-out order
  $H = ""; # the header of the file
  $H2 = ""; # body header
  $I = "\n<ul>"; # the index at the top of file
  $P = "\n<dl>"; # the body text of the file
  $Y = ""; # the footer

  $ws_name =  exists $wordset_name{$W} ? $wordset_name{$W} : "(unknown)";
  $ws_shortname =  exists $wordset_shortname{$W} ? $wordset_shortname{$W} : "";
  $H .= "\n<meta name=\"Author\" content=\"$ENV{USER}\">";
  $H .= "\n<title> Exported Names in Wordsets </title>\n";
  $H .= "\n".$style_nounder;

  $H2 .= "<h2>\n<!--HEADER-->\n $ws_name \n<!--header-->\n</h2>\n";
  
  do { open LOG, ">$LOG"; print STDERR "created $LOG\n"; } if $debug;

  my $ws_link = "<!--no wordset-->";
  if (length $ws_shortname) {
     $ws_link = "\n<!--wordset:--><nobr><tt>&lt;<a href=\"wordset-$W.html\">"
                   .v($ws_shortname)."</a>&gt;</tt></nobr>";
  }

  if (length $ws_shortname and exists $section_from_shortname{$ws_shortname}) {
     my $section_name = $section_from_shortname{$ws_shortname};
     $_ = $section_nohead{$section_name};
     s/\<\!/\< \!/sg;
     s/\-\>/\- \>/sg;
     $H2 .= "<!-- INFO from "; $H2 .= $section_file{$section_name};
     $H2 .= " -- \n"; $H2 .= $_; $H2 .= "* -->\n";
     $H2 .= "\n<p>\n"; 
     $H2 .= &replace_crossreferences($section_header{$section_name}); 
     $H2 .= "\n<p>\n";
  }

  foreach my $def (sort keys %wordset_a_symbol)
  {
     if ($def !~ /^$W:/) { next; }
     my $symbol = $wordset_a_symbol{$def};
     my $a_kind = $wordset_a_kind{$def};
     my $a_export = $wordset_a_export{$def};
     my $a_ansi = "";
     my $in_file;
     $a_ansi .= "<b> ansi $dpans{$a_export} </b>" if (exists $dpans{$a_export});

     
     print LOG "\n\ndef: $def $symbol $a_kind \"$a_export\"" if $debug;

     my $p = "\n<!--BODYTEXT-->\n<dl>"; # into export_name_file

     if (defined $fcode_comment{$symbol}) {
	my $head = $fcode_head{$symbol};
        my $body = $fcode_body{$symbol};
	$in_file = $fcode{$symbol};
        print LOG "\nsym: $symbol in $in_file with-comment" if $debug;
        $I .= "\n<li><a href=#$symbol>";
        $I .= &format_fcode_header($head, $a_export)."</a>";
        $P .= "\n<a name=$symbol>";

	$P .= "<dt>".&format_fcode_header($head, $a_export)."</dt></a>\n";
        $P .= "<dd><small> ".kind_name($a_kind).$a_ansi." </small><br>\n";
	$P .= &replace_crossreferences($body)."\n";

	$p .= "<dt>".&format_fcode_header($head, $a_export)."</dt></a>\n";
        $p .= "<dd><small> ".kind_name($a_kind).$a_ansi." </small>".$ws_link."<br>\n";
	$p .= &replace_crossreferences2($body)."\n";
     }elsif (exists $fcode{$symbol}) { 
        $in_file = $fcode{$symbol};
        print LOG "\nsym: $symbol in $in_file without-comment" if $debug;
        $I .= "\n<li><a href=#$symbol><bc>".v($a_export)."</bc> &nbsp; </a>";
        $I .= "<small> ".kind_name($a_kind)." </small>";

        $P .= "\n<a name=$symbol>";
        $P .= "<dt> <bc>".v($a_export)."</bc> </dt></a>\n";
        $P .= "<dd><small> ".&kind_name($a_kind).$a_ansi." </small><br>\n";
        $P .= "<p> no forth documentation available &nbsp; <small>($symbol)</small></p>\n";
	$P .= "<ul><li>see the ANSI Documentation of this word : <i>$a_ansi</i></li></ul>" if length $a_ansi;

        $p .= "<dt> <bc>".v($a_export)."</bc> </dt>\n";
        $p .= "<dd><small> ".&kind_name($a_kind).$a_ansi." </small>".$ws_link."<br>\n";
        $p .= "<p> no forth documentation available &nbsp; <small>($symbol)</small></p>\n";
	$p .= "<ul><li>see the ANSI Documentation of this word : <i>$a_ansi</i></li></ul>" if length $a_ansi;
	print STDERR "# no doc for forth $symbol ",v($a_export)," $a_ansi \n"
		    if exists $opt{"warn-nodoc"} and not length $a_ansi;
     }else{
        # no fcode definition found, but exported in wordset
        print LOG "\n???: '$symbol'" if $debug;
        $I .= "\n<li><a href=#$symbol><bc>".v($a_export)."</bc> &nbsp; </a>";
        $I .= "<small> ".kind_name($a_kind)." </small>";

        $P .= "\n<a name=$symbol><p></a>";
        $P .= "\n<dt> <bc>".v($a_export)."</bc> </dt>\n";
        $P .= "\n<dd><small> ".&kind_name($a_kind).$a_ansi." </small><br>\n";
        $P .= "<p> no documentation available &nbsp; <small>($symbol)</small></p>\n";
	$P .= "<ul><li>see the ANSI Documentation of this word : <i>$a_ansi</i></li></ul>" if length $a_ansi;

        $p .= "\n<dt> <bc>".v($a_export)."</bc> </dt>\n";
        $p .= "\n<dd><small> ".&kind_name($a_kind).$a_ansi." </small>".$ws_link."<br>\n";
        $p .= "<p> no documentation available &nbsp; <small>($symbol)</small></p>\n";
	$p .= "<ul><li>see the ANSI Documentation of this word : <i>$a_ansi</i></li></ul>" if length $a_ansi;
	print STDERR "# no doc for avail $symbol ",v($a_export)," $a_ansi \n"
			if exists $opt{"warn-nodoc"} or not length $a_ansi;

     }
     if (not defined $in_file) {
	$_ = $def; s/:.*//;
	$in_file = $wordset{$_};
        $_ = $def; s/.*://;
	$in_file .= ":".$_;
     }
     if (exists $opt{"refs"}) {
        $P .= "\n<p align=right><small> reference: ";
        $P .= "$symbol in $in_file, export $a_kind ".v($a_export);
        $P .= " </small></p>\n";

	$p .= "<hr>";
        $p .= "\n<p align=right><small> reference: ";
        $p .= "$symbol in $in_file, export $a_kind ".v($a_export);
        $p .= " </small></p>\n";
     }else{
	$P .= "\n<p><!-- reference: $symbol in $in_file, export $a_kind ".v($a_export)." -->\n";
	$p .= "\n<p><!-- reference: $symbol in $in_file, export $a_kind ".v($a_export)." -->\n";
     }
     $P .= "</dd>";
     $p .= "</dd></dl>";

     $p .= "\n<!--bodytext-->\n";
     $export_name_each{"$a_export"} = $p; undef $p;
  }

  $P .= "\n</dl>\n";
  $I .= "\n</ul>\n"; 

  close LOG if $debug;
  open F, ">$F";
  print STDERR "created $F\n";
  print F "<html><head>",
          "<meta name=\"date\" content=\"",scalar localtime(),"\">\n",
          "\n",o($H),
          "\n</head><body>\n",o($H2),
          "\n<hr>\n", o($I), "\n<hr>\n",
          "\n<!--BODYTEXT-->\n", o($P), "\n<!--bodytext-->\n", o($Y), 
          "\n</body></html>";
  close F;
}

# ............................. --each ................................

my @each_files;
if ($opt{"each"})
{
   my @out = sort keys %export_name_each;
   my $prev; my $curr;
   my $I = "<ul>\n";
   print STDERR "create file for EACH word: ";
   while (defined ($curr = shift @out))
   {
      my $FF = $export_name_file{$curr};
      if (defined $FF and open FF, ">$FF.html")
      {
  	 print STDERR "$FF.html ";
	 push @each_files, "$FF.html";
         print FF "<html><title>",v($curr),"</title><body>\n";
         print FF "\n<table width=100% bgcolor=blue border=0 cellspacing=1 cellpadding=5>",
                  "<tr><td><b><nobr><font color=white>";
         print FF "\n<!--HEADER-->\n",v($curr),"\n<!--header-->\n";
         print FF "</font></nobr></b></td><td align=center width=20%>\n";
	 print FF "<a href=\"",$export_name_file{$prev},".html\"><nobr><font color=white>",
		  "\n<!--PREV-->\n",v($prev),"\n<!--prev-->\n",
		  "</font></nobr></a>" if defined $prev;
         print FF "</td><td align=center width=20%>\n";
	 print FF "<a href=\"",$export_name_file{$out[0]},".html\"><nobr><font color=white>",
                  "\n<!--NEXT-->\n",v($out[0]),"\n<!--next-->\n",
                  "</font></nobr></a>" if defined $out[0];
         print FF "\n</td></tr></table>\n";
         print FF o($export_name_each{$curr}); # the body
         print FF "\n</body></html>\n";
         close FF;
         $I .= "<li><a href=\"$FF.html\"><c>".v($curr)."</c></a></li>\n";
      }
      $prev = $curr; 
   }
   {  my $FF = "index-each.htm";
      if (open FF, ">$FF") {        print STDERR "\ncreating $FF";
        print FF "<html><head><title>Index of each word</title></head>\n",
                "<meta name=\"date\" content=\"",scalar localtime(),"\">\n",
                "<body>\n",
                "<h2>Index of each word </h2>\n<hr>\n",
                o($I), 
                "\n</body></html>\n";
      }
      if ($opt{"each-index"} and length $opt{"each-index"}) {
	 my $FF = $opt{"each-index"};
	 if (open FF, ">$FF") { 
		print STDERR " (writing --each-index $FF ",$#each_files,")";
		print FF join ("\n",@each_files);
		close FF;
	 }
	 undef @each_files;
      }
   } print STDERR ".\n";
}

# .....................  src-*.html  ...........................

my %emitted;

if (exists $opt{"src-index"}) {
foreach my $file (@files)
{
  my $ws_name;
  my $file_basename = $file; 
  $file_basename =~ s(^.*/([^/]+)\s*$)($1) if $file_basename =~ m/\//;
  $file_basename =~ s(^(.*)\.(\w+)\s*$)($1) if $file_basename =~ m/\./;
  $F = "src-$file_basename.htm";
  $LOG = "src-$file_basename.log";
  # if you add more intermediate text-storages, keep the names
  # in alphabetical order with their spit-out order
  $H = ""; # the header of the file
  $H2 = ""; # body header
  $I = "\n<ul>"; # the index at the top of file
  $P = "\n<dl>"; # the body text of the file
  $Y = ""; # the footer

  $H .= "\n<meta name=\"Author\" content=\"$ENV{USER}\">";
  $H .= "\n<title> Source Definitions in $file_basename </title>\n";
  $H .= "\n".$style_nounder;

  $H2 .= "<h2> $file_basename definitions </h2>\n";
  
  do { open LOG, ">$LOG"; print STDERR "created $LOG\n"; } if $debug;

  foreach my $symbol (sort keys %fcode)
  {
     if ($fcode{$symbol} !~ /^$file:/) { next; }

     if ($ccode{$symbol})
     { # start ccode output

        $emitted{$symbol} = $F."l#".$symbol;
	if (defined $ccode_comment{$symbol})
	{
	    my $head = $ccode_head{$symbol};
	    my $body = $ccode_body{$symbol};
            $P .= "\n<a name=$symbol>";
	    $P .= "<dt><bc>$symbol :</bc>".v($head)."</dt></a>\n";
            $P .= "<dd><br>\n";
	    $P .= &replace_crossreferences($body)."\n";
            $I .= "\n<li><a href=#$symbol><bc>$symbol</bc></a>";
	}else{
            $P .= "\n<a name=$symbol>";
	    $P .= "<dt><bc>$symbol</bc></dt></a>\n";
            $P .= "<dd><br>\n";
            $I .= "\n<li><a href=#$symbol><bc>$symbol</bc></a>";
        }
        $P .= ((exists $opt{"refs"}) ?
		"\n<p align=right><small> reference: $ccode{$symbol} </small></p>\n"
		: "\n<!-- reference: $ccode{$symbol} -->\n" );
        $P .= "</dd>";
    } # done ccode output

    { # start fcode output

        $emitted{$symbol."_"} = $F."l#".$symbol."_";
	if (defined $fcode_comment{$symbol})
	{
	    my $head = $fcode_head{$symbol};
	    my $body = $fcode_body{$symbol};
            $P .= "\n<a name=$symbol\_>";
	    $P .= "<dt><bc>$symbol\_ :</bc> <c>".v($head)."</c></dt></a>\n";
            $P .= "<dd><br>\n";
	    $P .= &replace_crossreferences($body)."\n";
            $I .= "\n<li><a href=#$symbol\_><bc>$symbol\_</bc></a>";
	}else{
            $P .= "\n<a name=$symbol\_>";
	    $P .= "<dt><bc>$symbol\_ </bc> </dt></a>\n";
            $P .= "<dd><br>\n";
            $I .= "\n<li><a href=#$symbol\_><bc>$symbol\_</bc></a>";
	}
        $P .= ((exists $opt{"refs"}) ?
		"\n<p align=right><small> reference: $fcode{$symbol} </small></p>\n"
		: "\n<!-- reference: $fcode{$symbol} -->\n" );
        $P .= "</dd>";
     } # done fcode output
  } # done all fcode syms

  foreach my $symbol (sort keys %ccode)
  {
     if ($ccode{$symbol} !~ /^$file:/) { next; }
     if (exists $emitted{$symbol}) { next; }

     { # start ccode output

        $emitted{$symbol} = $F."l#".$symbol;
	if (defined $ccode_comment{$symbol})
	{
	    my $head = $ccode_head{$symbol};
	    my $body = $ccode_body{$symbol};
            $P .= "\n<a name=$symbol>";
	    $P .= "<dt><bc>$symbol :</bc>".v($head)."</dt></a>\n";
            $P .= "<dd><br>\n";
            $I .= "\n<li><a href=#$symbol><bc>$symbol</bc></a>";
	    $P .= &replace_crossreferences($body)."\n";
	}else{
            $P .= "\n<a name=$symbol>";
	    $P .= "<dt><bc>$symbol</bc></dt></a>\n";
            $P .= "<dd><br>\n";
            $I .= "\n<li><a href=#$symbol><bc>$symbol</bc></a>";
        }
        $P .= ((exists $opt{"refs"}) ?
		"\n<p align=right><small> reference: $ccode{$symbol} </small></p>\n"
		: "\n<!-- reference $ccode{$symbol} -->\n" );
        $P .= "</dd>";
     }
  }

  $P .= "\n</dl>\n";
  $I .= "\n</ul>\n"; 

  close LOG if $debug;
  open F, ">$F";
  print STDERR "created $F\n";
  print F "<html><head>",
          "<meta name=\"date\" content=\"",scalar localtime(),"\">\n",
          o($H),
          "\n</head><body>\n",o($H2),
          "\n<hr>\n", o($I), 
          "\n<hr>\n", o($P), o($Y), 
          "\n</body></html>";
  close F;
}} 

# .................... index-symbols.html .............................

if (exists $opt{"src-index"}) {
for my $F ($opt{"src-index"})
{
   $F = "index-symbols.htm" if not length $F;
   open F, ">$F" or next;
   print STDERR "created $F\n";

   print F "<html><head> <meta name=\"Author\" content=\"$ENV{USER}\">";
   print F "\n<title> Symbol Names in Source Files </title>\n";
   print F "\n", $style_nounder;
   print F "\n</head><body>";
   print F "\n";
   print F "\n<h2> Symbol Names </h2>";
   print F "\n";
   
   for my $W (sort keys %emitted)
   { 
	print F "<li> <a href=\"", $emitted{$W},"\">", 
                " <code>", v($W), " </code></a>";
      if (exists $ccode{$W})
      {
	print F "&nbsp; <small> ",$ccode{$W} , "</small>\n";
      }else{
        $W =~ s/_$//;
        if (exists $fcode{$W})
        {
        	print F "&nbsp; <small> ",$fcode{$W}, " </small>\n";
        }else{
	 	print F "\n";
        }
      }
   }
   print F "\n</body></html>\n";
   close F;	
}}

# .................... index-wordsets.html ..........................

{ my $P = "<table>\n";
  for my $ws (sort keys %wordset)
  {
    $P .= "<tr valign=top>\n";
    $P .= "<td align=right><a href=\"wordset-$ws.html\"> $ws </a></td>\n";
    $P .= "<td align=left><a href=\"wordset-$ws.html\">".v($wordset_name{$ws})."</a></td>\n";
    $P .= "</tr>\n";
  }
  $P .= "</table>\n";

  my $F = "index-wordsets.htm";
  if (open F, ">$F")
  {
    print STDERR "created $F\n";
    print F "<html><head> <meta name=\"Author\" content=\"$ENV{USER}\">";
    print F "\n<title> List Of Wordsets </title>\n";
    print F "\n", $style_nounder;
    print F "\n</head><body>";
    print F "\n";
    print F "\n<h2> List Of Wordsets </h2>";
    print F "\n";
    print F $P;
    print F "\n</body></html>\n";
    close F;	
  }
}
