/** @name CURSES-DRIVER --- curses-terminal low-level routines for term.c 
 *
 *  Copyright (C) Tektronix, Inc. 1998 - 2000. All rights reserved.
 *
 *  @see     GNU LGPL
 *  @author  Tektronix CTE              @(#) %derived_by: guidod %
 *  @version %version: 4.30 %
 *    (%date_modified: Thu Jan 27 21:58:35 2000 %)
 *
 *  @description
 *              Experimental terminal driver for UNIX-like systems
 *		using the curses library.
 *
 *		This driver works with Linux and the ncurses library.
 *		It doesn't work any better than the termcap.c driver.
 *		To use it change makefile as follows:
 * 
        	OPTS=	-I/usr/include/ncurses
                LIBS=	-lncurses
                TERM_O=	curses.o
 */
/*@{*/
#if defined(__version_control__) && defined(__GNUC__)
static char* id __attribute__((unused)) = 
      "@(#) $Id: %full_filespec: curses.c~4.30:csrc:bln_12xx!4 % $";
#endif

#define _P4_SOURCE 1

#include "p4sys.h"
#include "term.h"

#include <limits.h>
#include <sys/ioctl.h>		/* ioctl(), TIOCGWINSZ */

#if defined SYS_Linux
# include <ncurses.h>
#else
# include <curses.h>
#endif

#include "missing.h"

#define SEQ(X)	{ '\033', (p4char)((X) >> CHAR_BIT), (p4char)(X), 0 }

static p4char
rks [][4] =			/* what function keys send */
{
  SEQ (KEY_F(1)),	SEQ (KEY_F(2)),
  SEQ (KEY_F(3)),	SEQ (KEY_F(4)),
  SEQ (KEY_F(5)),	SEQ (KEY_F(6)),
  SEQ (KEY_F(7)),	SEQ (KEY_F(8)),
  SEQ (KEY_F(9)),	SEQ (KEY_F(10)),
  SEQ (KEY_LEFT),	SEQ (KEY_RIGHT),
  SEQ (KEY_UP),		SEQ (KEY_DOWN),
  SEQ (KEY_HOME),	SEQ (KEY_END),
  SEQ (KEY_PPAGE),	SEQ (KEY_NPAGE),
  SEQ (KEY_BACKSPACE),	SEQ (KEY_DC),
  SEQ (KEY_IC),		SEQ (KEY_IC),
  SEQ (KEY_IL),		SEQ (KEY_EOL),
  SEQ (KEY_DL),		SEQ (KEY_CLEAR)
};

char *term_curses_rawkey_string [] =
{
  rks [ 0], rks [ 1], rks [ 2], rks [ 3], rks [ 4],
  rks [ 5], rks [ 6], rks [ 7], rks [ 8], rks [ 9],
  rks [10], rks [11], rks [12], rks [13], rks [14],
  rks [15], rks [16], rks [17], rks [18], rks [19],
  rks [20], rks [21], rks [22], rks [23], rks [24],
  rks [25]
};

static int c_interrupt_key (char ch)		{ return 0; }
static void c_interactive_terminal (void)	{ fixterm (); }
static void c_system_terminal (void)		{ resetterm (); }

static int
c_prepare_terminal (void)
{
  initscr ();
  cbreak ();
  nonl ();
  noecho ();
  idlok (stdscr, TRUE);
  scrollok (stdscr, TRUE);
  meta (stdscr, TRUE);
  keypad (stdscr, TRUE);
  nodelay (stdscr, FALSE);
  saveterm ();
  refresh ();
  return 1;
}

static void
c_cleanup_terminal (void)
{
  return; /* nothing to do here */
}

#ifdef TIOCGWINSZ
static void
c_query_winsize (void)
{
  struct winsize size;

  if (ioctl (1, TIOCGWINSZ, (char *) &size) >= 0)
    {
      PFE.rows = size.ws_row;
      PFE.cols = size.ws_col;
      PFE.xmax = size.ws_xpixel;
      PFE.ymax = size.ws_ypixel;
    }
}

#else
static void c_query_winsize (void) {} 
#endif

static p4char nextch[3];
static int nnext = 0;

static int
key (void)
{
  int c;

  c = getch ();
  if (c == -1)
    return 0;
  if (c < 0x100)
    {
      nextch[nnext++] = c;
    }
  else
    {
      nextch[nnext++] = c;
      nextch[nnext++] = c >> CHAR_BIT;
      nextch[nnext++] = '\033';
    }
  return 1;
}

static int
c_keypressed (void)
{
  refresh ();
  if (nnext)
    return 1;
  nodelay (stdscr, TRUE);
  key ();
  nodelay (stdscr, FALSE);
  return nnext != 0;
}

static int
c_getkey (void)
{
  refresh ();

  if (PFE.wait_for_stdin)
    PFE.wait_for_stdin ();

  if (nnext || key ())
    return nextch[--nnext];
  else
    return -1;
}

static void 
c_putc_noflush (char c)
{
  addch (c);
}
static void c_put_flush ()		{ refresh (); }
static void c_putc (char c)		{ addch (c); refresh (); }
static void c_puts (const char *s)	{ addstr ((char *)s); refresh (); }

static void c_gotoxy (int x, int y)	{ move (y, x); }
static void c_wherexy (int *x, int *y)	{ getyx (stdscr, *y, *x); }

static void
addxy (int x, int y)
{
  int col, row;
  getyx (stdscr, row, col);
  move (row + y, col + x);
}

static void 
c_tput (int attr)
{
   switch (attr)
   {
   case P4_TERM_GOLEFT:		addxy (-1,  0); 	break;
   case P4_TERM_GORIGHT:	addxy ( 1,  0); 	break;
   case P4_TERM_GOUP:		addxy ( 0, -1); 	break;
   case P4_TERM_GODOWN:		addxy ( 0,  1); 	break;

   case P4_TERM_CLRSCR:		clear (); refresh (); 	break;
   case P4_TERM_HOME:		move (0, 0); 		break;
   case P4_TERM_CLREOL:		clrtoeol (); 		break;
   case P4_TERM_CLRDOWN:	clrtobot (); 		break;
   case P4_TERM_BELL:		beep (); 		break;
   
   case P4_TERM_NORMAL:		attrset (A_NORMAL); 	break;
   case P4_TERM_BOLD_ON:	standout (); 		break;
   case P4_TERM_BOLD_OFF:	standend (); 		break;
   case P4_TERM_BRIGHT:		attron (A_BOLD); 	break;
   case P4_TERM_REVERSE:	attron (A_REVERSE); 	break;
   case P4_TERM_BLINKING:	attron (A_BLINK); 	break;
   case P4_TERM_UNDERLINE_ON:	attron (A_UNDERLINE); 	break;
   case P4_TERM_UNDERLINE_OFF:	attroff (A_UNDERLINE);	break;

   default: break;
   }
}

#ifdef __GNUC__
#define INTO(x) .x =
#else
#define INTO(x)
#endif

p4_term_struct p4_term_ios =
{
  "curses",
  0, term_curses_rawkey_string,
  INTO(init) 		c_prepare_terminal, 
  INTO(fini) 		c_cleanup_terminal,
  INTO(tput)		c_tput,

  INTO(tty_interrupt_key) c_interrupt_key,
  INTO(interactive_terminal) c_interactive_terminal,
  INTO(system_terminal)   c_system_termainl,
  INTO(query_winsize)     c_query_winsize,

  INTO(c_keypressed)	c_keypressed,
  INTO(c_getkey)	c_getkey,
  INTO(c_putc_noflush)  c_putc_noflush,
  INTO(c_put_flush	c_put_flush,
  INTO(c_putc)		c_putc,
  INTO(c_puts)		c_puts,
  INTO(c_gotoxy)	c_gotoxy,
  INTO(c_wherexy)	c_wherexy
};



/*@}*/

