/** @name dictionary.c ---   Implements dictionary and wordlists.
 *
 *  Copyright (C) Tektronix, Inc. 1998 - 2000. All rights reserved.
 *
 *  @see     GNU LGPL
 *  @author  Tektronix CTE              @(#) %derived_by: guidod %
 *  @version %version: 4.49 %
 *    (%date_modified: Mon May 15 17:49:44 2000 %)
 */
/*@{*/
#if defined(__version_control__) && defined(__GNUC__)
static char* id __attribute__((unused)) = 
      "@(#) $Id: %full_filespec: dictnry.c~4.49:csrc:bln_12xx!4 %4 % $";
#endif

#define _P4_SOURCE 1

#include "p4sys.h"
#include "support.h"
#include "p4_comp.h"
#include "term.h"
#include "lined.h"

#include <string.h>
#include <ctype.h>
#include <stddef.h> /*offsetof*/
#include <stdlib.h>

#include "missing.h"
#include "p4_log.h"

/*
 * A vocabulary is organized as a mixture between hash-table and
 * linked list. (This is a practice you can observe in several
 * systems.) It works like this: Given a name, first a hash-code is
 * generated. This hash-code selects one of several linked lists
 * called threads. The hooks to these threads are stored in a table.
 *
 * The body of a WORDLIST is essentially such a table of pointers to
 * threads, while in FIG-Forth it was just a pointer to the one and
 * only linked list a VOCABULARY consists of in FIG-Forth.
 */

int
p4_wl_hash (const char *s, int l)
/* s string, l length of string, returns hash-code for that name */
{
#if 0 /*GD* original by DUZ */
  int n = *s++ - '@';

  while (--l > 0)
    n = n * 37 + *s++ - '@';	/* a maybe-stupid hash function :-) */
  return n & (THREADS - 1);	/* i.e. modulo threads */

#else /*GD* more simpler, avoiding multiply */
  register char c = *s;

  while(--l > 0)
    { c += *s++; c ^= l; }
  return c & (THREADS - 1);
#endif
}

/*
 * If we want to traverse a WORDLIST in it's entirety, we must follow
 * all threads simultaneously. The following definition eases this by
 * locating the thread with the hook pointing to the highest memory
 * location, assuming that this thread contains the latest definition
 * entered in the given WORDLIST. For usage refer to the definition of
 * WORDS.
 *
 * When following a wordlist using topmost, a copy of the word list
 * must be made. Everytime the topmost item was processed it must be
 * replaced by its successor in the linked list.
 */

char **
p4_topmost (Wordl *w)
/* find the thread with the latest word in the given word list */
{
  int n = THREADS;
  char **p, **s = w->thread;

  for (p = s++; --n; s++)
    if (*s > *p)
      p = s;
  return p;
}

char *				/* return the NFA of the latest */
p4_latest (void)		/* definition in the CURRENT WORDLIST */
{
  return *p4_topmost (CURRENT);
}

/* word list and forget */

Wordl *
p4_make_wordlist (void)
/* create a word list in the dictionary */
{
  Wordl *w = (Wordl *) DP;	/* allocate word list in HERE */
  P4_INC (DP, Wordl);

  ZERO (w->thread);		/* initialize all threads to empty */
  w->flag = WORDL_FLAG;         /* init flags from global flags */
  w->prev = VOC_LINK;		/* chain word list in VOC-LINK */
  VOC_LINK = w;
  if (w->flag & WORDL_CURRENT)
    w->also = CURRENT;		/* if WORDL_CURRENT, search also this */
  else
    w->also = 0;
  return w;
}

FCode (p4_forget_dp)
/* remove words from dictionary, free dictionary space */
{
    register Wordl *wl;
    register char* new_dp = PFE.forget_dp;

    /* unchain words in all threads of all word lists: */
    for (wl = VOC_LINK; wl; wl = wl->prev)
    {
        char **p = wl->thread;
        int i;
    
        for (i = THREADS; --i >= 0; p++)
	{  /* unchain words in thread: */
	    while (*p >= (char *) new_dp) 
            {
#	       ifdef USE_FFA
               if (*_FFA(*p) & P4xONxDESTROY) {
	            P4_info2 (" destroy: \"%.*s\"", **p, *p+1);
		    p4_call (p4_name_from (*p));
                    new_dp = PFE.forget_dp; /* forget_dp is volatile */
                      /* and may have changed through recursive forget */
	       }
#              endif
	       *p = *p4_name_to_link (*p);
            }
        }
    }

    /* unchain word lists: */
    while (VOC_LINK >= (Wordl *) new_dp) 
    {   
        {   /* delete from search-order */   
            int i;
            for (i=0; i < ORDER_LEN; i++) 
            {
                if (CONTEXT[i] == VOC_LINK) {
                   char* nfa = p4_to_name (BODY_FROM (VOC_LINK));
                   P4_warn3 ("deleted '%.*s' from context search-order [%i]", 
                  	    NFACNT(*nfa), nfa+1, i);
                   CONTEXT[i] = NULL;
                }
            
                if (PFE.dforder[i] == VOC_LINK) {
                   char* nfa = p4_to_name (BODY_FROM (VOC_LINK));
                   P4_warn3 ("deleted '%.*s' from default search-order [%i]", 
                  	    NFACNT(*nfa), nfa+1, i);
                   PFE.dforder[i] = NULL;
                }
            }
        }
        
        VOC_LINK = VOC_LINK->prev;
    }
    
    /* compact search-order */
    { register int i, j;
      for (i=0, j=0; i < ORDER_LEN; i++)
      {
        if (CONTEXT[i]) CONTEXT[j++] = CONTEXT[i];
      }
      while (j < ORDER_LEN) CONTEXT[j++] = NULL;

      for (i=0, j=0; i < ORDER_LEN; i++)
      {
        if (PFE.dforder[i]) PFE.dforder[j++] = PFE.dforder[i];
      }
      while (j < ORDER_LEN) PFE.dforder[j++] = NULL;
    }
    
    /* free dictionary space: */
    DP = (p4char *) new_dp; 
    LAST = NULL;
    PFE.forget_dp = 0;

    if (CURRENT >= (Wordl *) new_dp) 
    {
        if (CONTEXT[0]) CURRENT = CONTEXT[0]; /* implicit DEFINITIONS */
        p4_throw (THROW_CURRENT_DELETED);     /* but still throw */
    }
}

void
p4_forget (char* above)
{
   if ((p4char*) above < FENCE)
      p4_throw (THROW_INVALID_FORGET);

   if (PFE.forget_dp) /* some p4_forget_dp already started */
   {
      /* P4_info1 ("recursive forget %p", above); */
      if (PFE.forget_dp > above) 
      {
         PFE.forget_dp = above; /* update p4_forget_dp argument */
      }
   }else{ 
      /* P4_info1 ("forget start %p", above); */
      PFE.forget_dp = above; /* put new argument for p4_forget_dp */
      FX (p4_forget_dp);     /* forget execution start */
   }
}

Head *
p4_make_head (const char *name, int len, char **nfa, Wordl *wid);

char*
p4_forget_word (const char *name, p4cell id, p4code ccode, p4cell what)
{
  char* nfa;
  char nm[255];

  sprintf (nm, name, id);
  
  p4_make_head (nm, strlen(nm), &nfa, PFE.atexit_wl)->cfa  = ccode;
#ifdef USE_FFA
  *_FFA(nfa) |= P4xONxDESTROY;
#endif
  COMMA(what);
  return nfa;
}

void 
p4_forget_on_exit ()
{
   Wordl *wl;

   for (wl = VOC_LINK; wl; wl = wl->prev)
     {
        char**p = wl->thread;
        int i;
 
        for (i = THREADS; --i >= 0; p++)
	   while (*p)
	   {
#	      ifdef USE_FFA
		if (*_FFA(*p) & P4xONxDESTROY) {
			p4_call (p4_name_from (*p));
		}
#	      endif
	      *p = *p4_name_to_link (*p);
	   }
      }
    DP = 0;
    LAST = 0;
}

/* search a header */

static char *
search_thread (const char *nm, int l, char *t, p4cell wl_flag )
{
  if (l > NFACNTMAX)
    return NULL;

# if P4_LOG /* additional sanity check */
    if (p4_LogMask & P4_LOG_DEBUG) /* if any debug level */
        if (t && !((char*)PFE.dict <= t && t <= (char*)PFE.dictlimit)) 
        {	P4_fail3 ("hashlink pointer invalid %p in search for '%.*s'", 
                   	t, l, nm);
        }
# endif

  if( LOWER_CASE && (wl_flag & WORDL_NOCASE) )
  {
    /* this thread does only contain upper-case defs 
       AND lower-case input shall match definitions */
    while (t)
    {
      if ( !(*_FFA(t) & P4xSMUDGED) 
      && NFACNT(*t) == l && !strncmpi (nm, t+1, l))
	break;
      t = *p4_name_to_link (t);
    }
  }else{
    /* input is case-sensitive OR vocabulary contains mixed-case defs */
    while (t)
    {
      if ( !(*_FFA(t) & P4xSMUDGED) 
      && NFACNT(*t) == l && !strncmp (nm, t+1, l))
	break;
      t = *p4_name_to_link (t);
    }
  }

  return t;
}

char *
p4_search_wordlist (const char *nm, int l, /*const */ Wordl *w)
{
  if( w->flag & WORDL_NOHASH )
    { return search_thread (nm, l, w->thread[0], w->flag ); }
  else
    { return search_thread (nm, l, w->thread[p4_wl_hash (nm, l)], w->flag ); }
}

char *
p4_find (const char *nm, int l)
/* search all word lists in the search order for name, return NFA */
{
  Wordl **p, **q, *wordl;
  char *w = NULL;
  int n = p4_wl_hash (nm, l);

  for (p = CONTEXT; !w && p <= &ONLY; p++)
    {
      if (*p == NULL)
	continue;
      for (q = CONTEXT; *q != *p; q++); /* don't search wordl twice */
      if (q != p)
	continue;
      wordl = *p;
      do {
        if( wordl->flag & WORDL_NOHASH )
 	  w = search_thread (nm, l, wordl->thread[0], wordl->flag );
        else
          w = search_thread (nm, l, wordl->thread[n], wordl->flag );
      } while (!w && (wordl=wordl->also));
    }
  return w;
}

char *
p4_tick (p4xt *xt)		/* tick next word, store p4xt in xt, */
{				/* return count byte of name field */
  char *p;			/* (to detect immediacy) */

  p = p4_word (' ');
  p = p4_find ((char *) p + 1, *(p4char *) p);
  if (!p)
    p4_throw (THROW_UNDEFINED);
  *xt = p4_name_from (p);
  return p;
}

/* create a header */

char *
p4_alloc_string (const char *s, int len)
/* writes counted string into dictionary, returns address */
{
  char *p = (char *) DP;

  if (len >= (1 << CHAR_BIT))
    p4_throw (THROW_ARG_TYPE);
  *DP++ = len;			/* store count byte */
  while (--len >= 0)		/* store string */
    *DP++ = (p4char) *s++;
  FX (p4_align);
  return p;
}

char *
p4_alloc_parsed_string (char del)
{
  char *p;
  p4ucell n;

  p4_parse (del, &p, &n);
  return p4_alloc_string (p, (int) n);
}

char *
p4_alloc_word (char del)
{
  char *p = p4_word (del);

  DP += *p + 1;
  FX (p4_align);
  return p;
}

static void			/* written to cfa by make_head() */
illegal_xt (void)		/* to give an error msg when calling */
{				/* a word without execution semantics */
  p4_throw (THROW_INVALID_NAME);
}

Head *
p4_make_head (const char *name, int len, char **nfa, Wordl *wid)
/* make a new dictionary entry in the word list identified by wid */
{
  Head *h;
  int hc;

  if (len == 0)
    p4_throw (THROW_ZERO_NAME);
  if (len > NFACNTMAX)
    p4_throw (THROW_NAME_TOO_LONG);
  if (REDEFINED_MSG && p4_find (name, len))
    p4_outf ("\n\"%.*s\" is redefined ", len, name);
  /* p4_alloc_string .. */
  if (len >= (1 << CHAR_BIT))
    p4_throw (THROW_ARG_TYPE);
#ifdef USE_FFA
  *DP++ = '\x80'; 
  LAST = DP;
  *DP++ = len;
#else
  LAST = DP;
  *DP++ = len;			/* store count byte */
  *LAST |= 0x80; 	        /* LFA>NFA flag */
#endif
  { register int i=len;
    while (--i >= 0)		/* store string */
      *DP++ = (p4char) *name++;
  }

  FX (p4_align); 
  h = (Head *) DP;
  P4_INC (DP, Head);

  if (wid->flag & WORDL_NOHASH)
    hc = 0;
  else
    hc = p4_wl_hash (LAST+1, len); 
  h->link = wid->thread[hc];
  wid->thread[hc] = LAST;
  h->cfa = illegal_xt;
#ifndef USE_FIG
  h->aux = illegal_xt;
#endif

  if (nfa) *nfa = LAST;
  return h;
}

char* 
p4_header_from (p4code cfa, char flags, char* name, int count)
{
  char* nfa;
  p4_make_head (name, count, &nfa, CURRENT)->cfa = cfa;
  *_FFA(nfa) |= flags;
  return nfa;
}

char*
p4_header (p4code cfa, char flags)
{
  char* p; p4ucell len;
  p4_parseword (' ', &p, &len);
  return p4_header_from (cfa, flags, p, (int)len);
}

/* navigation in the header */

char **
p4_name_to_link (const char *p)
{
  return (char **) p4_aligned ((p4cell) (p+1 + NFACNT(*p)) );
}

char *
p4_link_to_name (char **l)
/*
 * scan backward for count byte preceeding name of definition
 * returns pointer to count byte of name field or NULL
 */
{
  char *p = (char *) l;
  unsigned n;

  /* Skip possible alignment padding: */
  for (n = 0; *--p == '\0'; n++)
    if (n > sizeof (p4cell) - 1)
      return NULL;

  /* Scan for count byte. Note: this is not reliable! */
  for (n = 0; n < NFACNTMAX; n++, p--)
    {
      if (*_FFA(p) & 0x80 && (unsigned)NFACNT(*p) == n)
	return p;
      if (!p4_isprint (*p))
	return NULL;
    }
  return NULL;
}

SemantP4 *			/* I don't like this either. :-) */
p4_to_semant (p4xt xt)
{
#define TO_SEMANT(XT,ELEMENT) \
  ((SemantP4 *)((char *)XT - OFFSET_OF (SemantP4, ELEMENT)))
  SemantP4 *s;

  s = TO_SEMANT (xt, exec[0]);
  if (s->magic == SEMANT_MAGIC)
    return s;
  s = TO_SEMANT (xt, exec[1]);
  if (s->magic == SEMANT_MAGIC)
    return s;
  return NULL;
#undef TO_SEMANT
}

p4xt
p4_link_from (char **lnk)
{
  return &((Head *)(void *) lnk)->cfa;  /*NOTE:* assumes lnk is first field*/
}

char **
p4_to_link (p4xt xt)
{
  SemantP4 *s = p4_to_semant (xt);

  return s ? p4_name_to_link (s->name)
    : (char**)( ((void*) xt) - offsetof(Head,cfa) ); 
                                	/*NOTE:* assumes lnk is first field*/
}

p4xt
p4_name_from (const char *p)
{
  return p4_link_from (p4_name_to_link (p));
}

char *
p4_to_name (p4xt c)
{
  return p4_link_to_name (p4_to_link (c));
}

p4xt
p4_runtime (void)
{
  if (!LAST)
    p4_throw (THROW_ARG_TYPE);
  return p4_name_from (LAST);
}

void
p4_dot_name (const char *nfa)
{
  if (!nfa || !(*_FFA(nfa) & 0x80))
    {
      p4_outs ("<\?\?\?> ");	/* avoid trigraph interpretation */
      return;
    }
  p4_type (nfa+1, NFACNT(*nfa));
  FX (p4_space);
}


/* words with wildcards */
void
p4_wild_words (const Wordl *wl, const char *pattern, const char *categories)
/*
 * Show words in word list matching pattern, and of one of the
 * categories in string `categories'. NULL pointer or zero length
 * string means all kinds of words.
 */
{
  Wordl wcopy = *wl;		/* clobbered while following it */
  char **t;

# ifndef WILD_TAB
# define WILD_TAB 26 /* traditional would be 20 */
# endif

  FX (p4_cr);
  FX (p4_start_Q_cr);
  if (categories && *categories == '\0')
    categories = NULL;
  for (t = p4_topmost (&wcopy); *t; t = p4_topmost (&wcopy))
    {
      char wbuf[NFACNTMAX+1];
      char *w = *t;
      char **s = p4_name_to_link (w);
      int l = NFACNT(*w++);

      p4_store_c_string (w, l, wbuf, sizeof wbuf);
      if (p4_match (pattern, wbuf, wl->flag & WORDL_NOCASE))
	{
	  char c = p4_category (*p4_link_from (s));
	  if (!categories || strchr (categories, c))
	    {
	      if (OUT + WILD_TAB - OUT % WILD_TAB + 2 + l > PFE.cols)
		{
		  if (p4_Q_cr ())
		    break;
		}
	      else
		{
		  if (OUT)
		    p4_tab (WILD_TAB);
		}
	      p4_outf ("%c %.*s ", c, l, w);
	    }
	}
      *t = *s;
    }
}

/* completion of word against dictionary */

static char *
search_incomplete (const char *name, int len, Wordl *w)
/*
 * traverses the entire given wordlist to find a matching word
 * caution: clobbers *w. This is needed to be able to continue the search.
 */
{
  char **t, *s;

  for (t = p4_topmost (w); *t; t = p4_topmost (w))
    {
      s = *t;
      *t = *p4_name_to_link (*t);
      if (NFACNT(*s) >= len 
      && ( (w->flag & WORDL_NOCASE) ? strncmpi (s + 1, name, len) 
				    : strncmp  (s + 1, name, len) 
         ) == 0)
	return s;
    }
  return NULL;
}

static int
p4_complete_word (const char *in, int len, char *out, int display)
/*
 * Try to complete string in/len from dictionary.
 * Store completion in out (asciiz), return number of possible completions.
 * If display is true, display alternatives.
 * 29JUN99 gd: if (display && !len) don't print 200 words, just the number
 */
{
  Wordl w, *wl, **p;
  char *s = NULL, *t = NULL;	
  int n = 0, m = 0, cnt = 0, searched_n = 0;
  Wordl* searched[32] = {0}; 
  
  for (p = CONTEXT; p <= &ONLY; p++)
  {
     for (wl = *p; wl; wl = wl->also ) 
     {
        for (n=0; n < searched_n; n++)
           if (wl == searched[n])
              break;                      /* continue at second outer for */
        if (wl == searched[n]) continue;  /* must expressed like that in C*/
        
        if (searched_n < 32) 
            searched[searched_n++] = wl;
             
        for (w = *wl; (t = search_incomplete (in, len, &w)) != NULL; cnt++)
	{
	   if (display && len) {
	      FX (p4_space);
	      p4_type_on_line (t + 1, NFACNT(*t));
	   }
	   if (cnt == 0) {
	      s = t + 1;
	      m = NFACNT(*t);
	   }else{
	      ++t;
	      for (n = 0; n < m; n++)
		if (s[n] != t[n])
		  break;
	      m = n;
	   }
	}
     }
  }
  if (cnt)
    p4_store_c_string (s, m, out, NFACNTMAX+1);
  if (display && !len)
     { p4_outf (" %i words ", cnt); }
  return cnt;
}

int
p4_complete_dictionary (char *in, char *out, int display)
{
  char *lw, buf[NFACNTMAX+1];
  int n;

  lw = strrchr (in, ' ');
  if (lw)
    lw++;
  else
    lw = in;
  memcpy (out, in, lw - in);
  n = p4_complete_word (lw, strlen (lw), buf, display);
  strcpy (&out[lw - in], buf);
  return n;
}

/************************************************************************/
/* initial dictionary setup                                             */
/************************************************************************/


char*
p4_make_word ( char type, const char* name, int len, void* ptr, Wordl *wid)
{
  Head *h;
  char *nfa;
  
  if (type == _CX || type == _CY) {
     wid = PFE.atexit_wl;
  }

  h = p4_make_head (name, len, &nfa, wid);
  if (type & 040)
    *_FFA(nfa) |= P4xIMMEDIATE;
  switch (type)
  {
    case _CS:
#       ifdef USE_FIG
          /* well, there's nothing good with this solution... *FIXME*/
          if (!PFE_FIG) PFE_FIG = calloc(1024,sizeof(void*));
          { register void** fig = PFE_FIG;
            while (*fig) fig++;
            *fig = ptr;
          }
#	else
	  h->aux = ptr;
#	endif
	  h->cfa = ((SemantP4 *) ptr)->comp;
	  ((SemantP4 *) ptr)->name = nfa;
	  return nfa;
    case _CI:
    case _CO:
	  h->cfa = (p4code) ptr;
	  return nfa;
    case _CX:
	  h->cfa = (p4code) ptr;
	  ((p4code)ptr) ();     /* runs *now* !! */
	  return nfa;
    case _CY:
	  h->cfa = (p4code) ptr;
#         ifdef USE_FFA
	    *_FFA(nfa) |= P4xONxDESTROY;
#         endif
	  return nfa;
/*
    case _VO:
	  h->cfa = p4_vocabulary_RT_ ;
	  (((WList*) ptr)->wid = p4_make_wordlist ()); 
          return nfa;
*/
    case _SV:
	  h->cfa = p4_sysvar_RT_ ;
	  break;
    case _DV:
	  h->cfa = p4_dictvar_RT_ ;
	  break;
    case _DC:
	  h->cfa = p4_dictconst_RT_ ;
	  break;
    case _SC:
	  h->cfa = p4_sysconst_RT_ ;
	  break;
    case _OV:
    case _IV:
	  h->cfa = p4_create_RT_ ;
	  break;
    case _OC:
    case _IC:
	  h->cfa = p4_constant_RT_ ;
	  break;
    case _OL:
    case _IL:
	  h->cfa = p4_value_RT_ ;
	  break;
    case _OT:
          h->cfa = p4_offset_RT_ ;
          break;
  }
  COMMA (ptr);
  return nfa;
}

FCode (p4_only_runtime)
{
  ZERO (CONTEXT);
  CONTEXT[0] = CURRENT = ONLY;
}

FCode (p4_forget_wordset_RT)
{
  P4_info1 ("dump '%s'", (*(p4Words**)WP_PFA)->name); 

  /* do nothing so far, forget_wordset_RT_ is just a type-marker */
}

extern int p4_slot_use (int*); /* FIXME: move to header file ? */
extern int p4_slot_unuse (int*); /* FIXME: move to header file ? */

FCode (p4_forget_slot_RT)
{
   int* slot = (int*)(WP_PFA[0]);
   P4_info1 ("unuse load-slot '%i'", *slot);
   
   if (slot && *slot && PFE.p[*slot]) {
      p4_xfree (PFE.p[*slot]); PFE.p[*slot] = 0;
   }
   
   p4_slot_unuse (slot);
}

void 
p4_load_llist (void** list, Wordl* wid, int extending)
{
   int e;
   void* p;
   Wordl* save_current = CURRENT;
   if (!list) return;
   
   while (*list) 
   { 
     /* switch(void*) P4_LOAD_XXXX */     
     if (*list == P4_LOAD_INTO) 
     {    register char* vocname = *(char**)(++list);

   	  /* 1. lookup Wordl, or create one if required */
          
          if (!extending)
	    p = 0;
          else 
          { /* check if there is already a vocabulary of that name,
	       in which case we just want to add another few words to it
            */
            p = p4_search_wordlist (vocname, strlen (vocname), wid);
            if (p) 
            { p = p4_name_from ((char*) p);
	      if (*((p4xt)p) != &p4_vocabulary_RT_ )
		    p = 0;
            }
          }
  
          if (p) 
          {   CURRENT = (Wordl*)TO_BODY ((p4xt) p);
              P4_info1 ("load into old '%s'", vocname);
          } else {
              CURRENT = wid;
              p4_header_from (p4_vocabulary_RT_ , P4xIMMEDIATE,
                                	vocname, strlen(vocname));
                 
	      CURRENT = p4_make_wordlist (); 
              P4_info1 ("load into new '%s'", vocname);
          }
          
          /* 2. load Wordsets into Wordl */
          
          ++list;
          for (; *list < 0 || *list > P4_LOAD_MAX; ++list)
          {
               int k = (*(p4Words**)list)->n;
               const p4Word* w = (*(p4Words**)list)->w;
               char dictname[32]; char* dn;
               
               if ((*(p4Words**)list)->name) 
               {  P4_info1 ("load '%s'", (*(p4Words**)list)->name);
                  strncpy (dictname, (*(p4Words**)list)->name, 31);
                  dictname[31] = '\0';
                  if ((dn= strchr (dictname, ' '))
                  ||  (dn= strchr (dictname, '(')))
                     *dn = '\0';
               } else {
                  sprintf (dictname, "%p", DP);
               }
               
               p4_forget_word ("wordset:%s", (p4cell) dictname,
                  	     p4_forget_wordset_RT_, 
                                	(p4cell) (*(p4Words**)list));
               
               for ( ; --k >= 0; w++)
               {
                 if (!w) continue;
                 p4_make_word (*w->name, w->name+1, strlen (w->name+1), 
                        	w->ptr, CURRENT);
               }
          }
          
     } else
     if (*list == P4_LOAD_EXEC) 
     {    register void (*fx)() = *(void**)(++list);
          
          P4_info1 ("load exec %p", fx);
          if (fx) {fx();}
     } else
     if (*list == P4_LOAD_ORDER) 
     { 	  register char* vocname = *(void**)(++list);
          
          Wordl* voc = p4_find_vocabulary (vocname, strlen(vocname));
          if (voc) 
          {
             { register int i;
               for (i=ORDER_LEN; --i > 0; )
                  if (CONTEXT[i] == voc) 
                  {
                     P4_info1 ("search also '%s' : already there", vocname);
                     goto break_P4_LOAD_ORDER;
                  }
             }
             FX (p4_also);    /* the top-of-order (CONTEXT) isn't changed */
             CONTEXT [1] = voc; /* instead we place it under-the-top */
             P4_info1 ("search also '%s' : done", vocname);
          } else {
             P4_warn2 ("search also failed: no '%s' vocabulary (%i)", 
                	vocname, strlen(vocname));
          }
          break_P4_LOAD_ORDER: ;
     } else         
     if (*list == P4_LOAD_SLOT) 
     {   register int* slot = *(int**)(++list);
         register int  size = (int) *(void**)(++list);
         
         if (!slot || size < 4)
                continue;
                
         if ((e=p4_slot_use (slot))) {
               P4_fail2 ("load-slot %i failed : %s", *slot, strerror(-e));
               break; /* return !!!! */
         }

         if (!(PFE.p)[*slot]) {
               (PFE.p)[*slot] = p4_calloc (1, size);
               P4_info3 ("load-slot %i size %i alloc (%p)", 
                  	  *slot, size, (PFE.p)[*slot]);
         } else 
               P4_warn2 ("load-slot %i already allocated (%p)", 
                  	  *slot, (PFE.p)[*slot]);

   	 p4_forget_word ("(load-slot: %i)", *slot, 
                	p4_forget_slot_RT_, (p4cell) slot);
     } else         
     {
          list++;
     } /* switch LOAD_COMMAND */
   } /* while not LOAD_END */

   CURRENT = save_current;
}

void
p4_preload_only (void)
{
  Wordl only;                   /* scratch ONLY word list */
  char* nfa;

  DP = (p4char *) &PFE.dict[1];

  /* Load the ONLY word list to the scratch ONLY: */
  memset (&only, 0, sizeof only);
  p4_make_head ("ONLY", 4, &nfa, &only )->cfa = p4_only_RT_;
  ONLY = p4_make_wordlist ();
  COPY (ONLY->thread, only.thread);   /* Copy scratch ONLY to real ONLY */
  
  CURRENT = ONLY;
  p4_header_from (p4_vocabulary_RT_ , P4xIMMEDIATE, "LOADED", 6);
  PFE.atexit_wl = p4_make_wordlist (); 
  PFE.atexit_wl->flag |= WORDL_NOHASH; /* see environment_dump in core.c */
}

/*@}*/

