/** @name LOCALS --- The Optional Locals Word Set
 * 
 *  Copyright (C) Tektronix, Inc. 1998 - 2000. All rights reserved.
 *
 *  @see     GNU LGPL
 *  @author  Tektronix CTE              @(#) %derived_by: guidod %
 *  @version %version: 4.31 %
 *    (%date_modified: Tue Apr 11 15:07:33 2000 %)
 *
 *  @description
 *      The Portable Forth Environment does implement locals
 *      in such an extended form as that additional variable
 *      names can be declared anywhere in the compiled word.
 */
/*@{*/
#if defined(__version_control__) && defined(__GNUC__)
static char* id __attribute__((unused)) = 
      "@(#) $Id: %full_filespec: locals.c~4.31:csrc:bln_12xx!4 % $";
#endif

#define _P4_SOURCE 1
#include "p4sys.h"

#include <string.h>

#include "support.h"
#include "p4_comp.h"

/** Did change the locals-implementation completely.
  Now the return stack looks like this after locals_bar_execution:

  rp --> 'rp' -- contains real return used by locals_exit_execution
  lp --> 'lp' -- contains locals-pointer of surrounding scope
          arg1
           ..
          argN
          localN+1
            ..
          localM
  'rp' --> -- real return address

   you can declare now a local non-arg at any time.
   All local-args must be declared before any local-non-arg.
   Not that the code after locals_bar_execution does not contain 
   _two_ numbers N and M. Classic use of LOCALS| will put them
   at the very same value.
*/   

/* 1. Actions at runtime: */

FCode (p4_local_enter_execution) /* establish local variables on return stack */
{
  p4cell *Rp = FX_RP;

  Rp = FX_RP - P4_POP(IP);
  P4_POSH(Rp) = (p4cell) LP;  LP = Rp;
  P4_POSH(Rp) = (p4cell) RP;  RP = (p4xt**) Rp;
}
/* P4_SKIPS_CELL + x */

FCode (p4_local_args)
{
  memcpy (LP+1, SP, (*(p4cell*)IP) * sizeof(p4cell));
  SP += P4_POP(IP);
} 
/* P4_SKIPS_CELL */

FCode (p4_locals_bar_execution)
{
  FX (p4_local_enter_execution);
  FX (p4_local_args);
}
/* P4_SKIPS_DCELL + x */

FCode (p4_locals_exit_execution)	/* alternative EXIT */
{				/* cleans up local variable stack frame */
  LP = (p4cell *) RP[1];
  RP = (p4xt **) *RP;
  IP = *RP++;
}

FCode (p4_local_execution)		/* retrieve current value of local variable */
{
  P4_POSH(SP) = LP[P4_POP(IP)];     /*  had been: LP[P4_POP(*IP)]; */
}

FCode (p4_to_local_execution)	/* set current value of local variable */
{
  LP[P4_POP(IP)] = P4_POP(SP);
}

/* 2. Actions at compile time */

int
p4_find_local (char *nm, int l)	/* returns index i to access local variable */
{				/* relative to LP [i], 0 if not defined */
  int i;
  int (*cmp)(const char*, const char*, int) = 
     	(int (*)(const char*, const char*,int))strncmp;

  if (LOWER_CASE)
    cmp = (int (*)(const char*, const char*,int))strncmpi;
  for (i = 0; i < *PFE.locals; i++)
    if ((*cmp) (nm, PFE.local[i], l) == 0 && PFE.local[i][l] == 0)
      return i + 1;
  return 0;
}

int
p4_compile_local (char *name, int len)
{
  static p4code cfa = PFX (p4_local_execution);
  int n;

  if ((n = p4_find_local (name, len)) == 0)
    return 0;
  COMMA (&cfa);
  COMMA (n);
  return 1;
}

static void 
enter_locals()
{
  PFE.locals = (p4cell *) DP;
  COMMA (0);
  COMMA (0);
}

static void
paren_local (char *nm, int l)
{
  FX (p4_Q_comp);
  if (l == 0)
    return;
  if (l > NFACNTMAX ) 
    p4_throw (THROW_NAME_TOO_LONG);
  if (PFE.locals == NULL) {
      p4_compile1 ();
      enter_locals ();
  }
  else{
    if (p4_find_local (nm, l)) { 
       p4_hereword (nm, l);
       p4_throw (THROW_INVALID_NAME);
    }
  }
  p4_store_c_string (nm, l, PFE.local[(*PFE.locals)++], NFACNTMAX+1);
}

/** ( strptr strcnt -- )
 * this word is used to create compiling words that can
 * declare => LOCALS| - it shall not be used directly
 * to declare a local, the pfe provides => LVALUE for
 * that a purpose beyond => LOCALS| 
 */
FCode (p4_paren_local)
{
  paren_local ((char *) SP[1], SP[0]);
  SP += 2;
  PFE.locals[1] ++;
}
P4COMPILES (p4_paren_local, p4_locals_bar_execution,
	  P4_SKIPS_DCELL, P4_LOCALS_STYLE);

/** LOCALS| name1 name2 ... nameN | ( xN ... x2 x1 -- )
 * create local identifiers to be used in the current definition.
 * At runtime, each identifier will be assigned a value from
 * the parameter stack. <br>
 * The identifiers may be treated as if being a => VALUE , it does
 * also implement the ansi => TO extensions for locals. Note that
 * the identifiers are only valid inside the currently compiled
 * word, the => SEE decompiled word will show them as 
 * => <A> => <B> ... => <N> a.s.o.   <br>
 * see also => LVALUE
 */
FCode (p4_locals_bar)
{
  for (;;)
    {
      char *p; int l;
      p4_parseword (' ', &p, &l);

      if (l == 1 && *p == '|')
	break;
      paren_local (p, l);
      PFE.locals[1] ++;
    }
}
P4COMPILES (p4_locals_bar, p4_locals_bar_execution,
	  P4_SKIPS_DCELL, P4_LOCALS_STYLE);

/** LVALUE name ( value -- )
 * declares a single local => VALUE using => (LOCAL) - a 
 * sequence of => LVALUE declarations can replace a 
 * => LOCALS| argument, ie. <c> LOCALS| a b c | </c> 
 * is the same as <c> LVALUE a  LVALUE b  LVALUE c </c>.
 * This should also clarify the runtime stack behaviour of
 * => LOCALS| where the stack parameters seem to be
 * assigned in reverse order as opposed to their textual
 * identifier declarations. <br>
 * compare with => VALUE and the pfe's convenience word
 * =>'VAR'.
 */
FCode (p4_local_value)
{
  char* nm; int l;
  if (!PFE.locals) {   /* need to build a dummy (LOCAL) entry */
    COMPILE1 (p4_paren_local); 
    enter_locals ();
  }
  p4_parseword (' ', &nm, &l);
  paren_local (nm, l);
  p4_compile2 ();
  COMMA (p4_find_local (nm,l));
}

P4COMPILES2(p4_local_value, p4_to_execution, p4_to_local_execution, 
		P4_SKIPS_TO_TOKEN, P4_LOCALS_STYLE);

LISTWORDS (locals) =
{
  CS ("(LOCAL)", p4_paren_local),
  CS ("LOCALS|", p4_locals_bar),
  CS ("LVALUE",  p4_local_value),
};
COUNTWORDS (locals, "LOCALS-EXT Locals + extensions");

/*@}*/

