/** @name FORTH-83-L&P --- Compatibility with Laxen&Perry's F83.
 *
 *  Copyright (C) Tektronix, Inc. 1998 - 2000. All rights reserved.
 *
 *  @see     GNU LGPL
 *  @author  Tektronix CTE               @(#) %derived_by: guidod %
 *  @version %version: 4.35 %
 *    (%date_modified: Thu Jan 27 20:18:37 2000 %)
 *
 *  @description
 *              There are lots of useful words in F83 that do not appear
 *              in any standard. This wordset defines some of them.
 */
/*@{*/
#if defined(__version_control__) && defined(__GNUC__)
static char* id __attribute__((unused)) = 
      "@(#) $Id: %full_filespec: lpf83.c~4.35:csrc:bln_12xx!4 %4 % $";
#endif

#define _P4_SOURCE 1

#include "p4sys.h"
#include "support.h"
#include "p4_comp.h"

#include <string.h>
#include <ctype.h>

#include "missing.h"
#include "p4_log.h"

/************************************************************************/
/* from KERNEL86.BLK                                                    */
/************************************************************************/

/** BOUNDS ( a b -- b+a a )
 simulate:
   : BOUNDS  OVER + SWAP ; 
 */
FCode (p4_bounds)		
{
  p4cell h = SP[1];

  SP[1] += SP[0];
  SP[0] = h;
}

/** PERFORM ( addr -- ? ) 
  simulate:
    : PERFORM  @ EXECUTE 
  */
FCode (p4_perform)	
{		
  PFE.execute (*(p4xt *) *SP++);
}

/** ?LEAVE ( cond -- )
 * leave a (innermost) loop if condition is true
 */
FCode (p4_question_leave)
{
  if (*SP++)
    {
      IP = RP[2] - 1;
      RP += 3;
      FX_BRANCH;
    }
}

/** NOOP ( -- )
 * do nothing, used as a place-holder where 
 * an execution word is needed
 */
FCode (p4_noop)
{
}

/** RP@ ( -- addr ) 
 * returns the return stack pointer 
 example:
   : R@ RP@ @ ;
 */
FCode (p4_r_p_fetch)		
{				
  *--SP = (p4cell) RP;
}

/** RP! ( addr -- ) 
 * sets the return stack pointer, reverse of => RP@
 */
FCode (p4_r_p_store)		
{				
  RP = (p4xt **) *SP++;
}

/** SP! ( ... addr -- ) 
 * sets the stack pointer, reverse of => SP@
 */
FCode (p4_s_p_store)		
{				
  SP = *(p4cell **) SP;
}

/** -ROT ( a b c -- c a b )
 * inverse of => ROT
 */
FCode (p4_dash_rot)	
{
  p4cell h = SP[2];

  SP[2] = SP[0];
  SP[0] = SP[1];
  SP[1] = h;
}

/** CSET ( n addr -- ) 
 * set bits in byte at given address 
 simulate:
   : CSET  TUCK @ SWAP OR SWAP ! ;
 */
FCode (p4_c_set)			
{			
  *(char *) SP[0] |= (char) SP[1];
  SP += 2;
}

/** CRESET ( n addr -- ) 
 *  reset bits in byte at given address 
 simulate:
   : CRESET  TUCK @ SWAP NOT AND SWAP ! ;
 */
FCode (p4_c_reset)			
{				
  *(char *) SP[0] &= ~(char) SP[1];
  SP += 2;
}

/** CTOGGLE ( n addr -- ) 
 * toggle bits in byte at given address 
 simulate:
   : CTOGGLE  TUCK @ SWAP XOR SWAP ! ;
 */
FCode (p4_c_toggle)			
{				
  *(char *) SP[0] ^= (char) SP[1];
  SP += 2;
}

/** OFF ( addr -- )
 simulate:
   : OFF  FALSE SWAP ! ;
 */
FCode (p4_off)			
{
  *(p4cell *) *SP++ = P4_FALSE;
}

/** ON  ( addr -- )
 * simulate:
   : ON  TRUE SWAP ! ;
 */
FCode (p4_on)			
{
  *(p4cell *) *SP++ = P4_TRUE;
}

/** 3DUP ( a b c -- a b c a b c )
 simulate:
   : 3DUP  3 PICK 3 PICK 3 PICK ;
 */
FCode (p4_three_dup)	
{
  SP -= 3;
  SP[0] = SP[3];
  SP[1] = SP[4];
  SP[2] = SP[5];
}

/** 4DUP ( a b c d -- a b c d a b c d )
 simulate:
  : 4DUP  4 PICK 4 PICK 4 PICK 4 PICK ;
 */
FCode (p4_four_dup)			
{
  SP -= 4;
  SP[0] = SP[4];
  SP[1] = SP[5];
  SP[2] = SP[6];
  SP[3] = SP[7];
}

/** UPC ( c1 -- c2 ) 
 * convert a single character to upper case 
   : UPC  >R _toupper ;
 */
FCode (p4_upc)			
{				
  *SP = toupper (*SP);
}

/** UPPER ( addr cnt -- ) 
 * convert string to upper case 
 simulate:
   : UPPER  0 DO  DUP I +  DUP C@ UPC SWAP C!  LOOP  DROP ;
 */
FCode (p4_upper)			
{				
  p4_upper ((char *) SP[1], SP[0]);
  FX_2DROP;
}

/** LOWER ( addr cnt -- ) 
 * convert string to lower case
 * This is not in L&P's F83 but provided for symmetry 
 simulate:
   : LOWER  0 DO  DUP I +  DUP C@ >R _tolower SWAP C!  LOOP  DROP ;
 */
FCode (p4_lower)			
{				
  p4_lower ((char *) SP[1], SP[0]);
  FX_2DROP;
}

/** SKIP ( addr cnt c -- addr' cnt' ) 
 * skip leading characters c 
 simulate:
   : SKIP  >R BEGIN DUP OVER C@ R@ = OR WHILE 
                    1- SWAP 1- SWAP REPEAT R> DROP ;
 */
FCode (p4_skip)			
{			
  char *p = (char *) SP[2];
  p4cell n = SP[1];
  char c = (char) *SP++;

  while (n && *p == c)
    n--, p++;
  SP[1] = (p4cell) p;
  SP[0] = n;
}

/** SCAN ( addr cnt c -- addr' cnt' ) 
 * scan for first occurence of c in string 
 simulate:
   : SCAN >R BEGIN DUP OVER C@ R@ = 0= OR WHILE 
                    1- SWAP 1- SWAP REPEAT R> DROP ;
 */
FCode (p4_scan)			
{				
  char *p = (char *) SP[2];
  p4cell n = SP[1];
  char c = (char) *SP++;

  while (n && *p != c)
    n--, p++;
  SP[1] = (p4cell) p;
  SP[0] = n;
}

/** PLACE ( addr1 len addr2 -- ) 
 * store string addr1/len at addr2 
 simulate:
   : PLACE  2DUP C! 1+ SWAP CMOVE ;
 */
FCode (p4_place)		
{				
  p4char *p = (p4char *) SP[0];

  *p = SP[1];
  memcpy (p + 1, (p4char *) SP[2], *p);
  SP += 3;
}

/** ASCII word ( -- val )
 * state smart version of => CHAR or => [CHAR] resp.
 simulate:
   : ASCII  [COMPILE] [CHAR] 
            STATE @ IF [COMPILE] LITERAL THEN ;
 */
FCode (p4_ascii)
{
  char *p;
  p4ucell n;

  p4_skip_delimiter (' ');
  p4_parse (' ', &p, &n);
  if (n == 0)
    p4_throw (THROW_INVALID_NAME);
  if (STATE)
    {
      p4_compile1 ();
      FX_COMMA (*(p4char *) p);
    }
  else
    FX_PUSH ( *(p4char *) p);
}
P4COMPILES (p4_ascii, p4_literal_execution,
	  P4_SKIPS_CELL, P4_DEFAULT_STYLE);

/** CONTROL word ( -- val )
 * see =>'ASCII', but returns char - '@' 
 simulate:
   : CONTROL  [COMPILE] [CHAR]  [CHAR] @ -  
              STATE @ IF [COMPILE] LITERAL THEN ;
 */
FCode (p4_control)
{
  char *p;
  p4ucell c;
  p4ucell n;

  p4_skip_delimiter (' ');
  p4_parse (' ', &p, &n);
  if (n == 0)
    p4_throw (THROW_INVALID_NAME);
  c = *(p4char *) p;
  if ('@' <= c && c <= '_')
    c -= '@';
  if (STATE)
    {
      p4_compile1 ();
      FX_COMMA (c);
    }
  else
    FX_PUSH (c);
}
P4COMPILES (p4_control, p4_literal_execution,
	  P4_SKIPS_CELL, P4_DEFAULT_STYLE);

/** NUMBER? ( addr -- d flag ) 
 * convert counted string to number - used in inner interpreter 
 * ( => INTERPRET ), flags if conversion was successful
 example:
   BL WORD  HERE NUMBER? 0= IF ." not a number " THEN . 
 */
FCode (p4_number_question)	
{			
  char *p = (char *) *SP;

  SP -= 2;
  SP[0] = p4_number_question (p + 1, *(p4char *) p, (p4dcell *) &SP[1]);
}

/************************************************************************/
/* from EXTEND86.BLK                                                    */
/************************************************************************/

/** VOCS ( -- )
 * list all vocabularies in the system
 simulate:
   : VOCS VOC-LINK @ BEGIN DUP WHILE
                           DUP BODY> >NAME .NAME
                           ->VOC-LINK @
                     REPEAT DROP ; 
 */
FCode (p4_vocs)	
{
  Wordl *wl = VOC_LINK;

  while (wl != NULL)
    {
      p4_dot_name (p4_to_name (BODY_FROM (wl)));
      wl = wl->prev;
    }
}

/** ((DEFER)) ( -- )
 * runtime of => DEFER words
 */
FCode (p4_defer_RT)
{
  p4xt xt = (p4xt) *WP_PFA;
  if (xt) { PFE.execute (xt); }
  else    { P4_warn ("null execution in deferred word"); }
}

/** DEFER word ( -- )
 * create a new word with ((DEFER))-semantics
 simulate:
   : DEFER  CREATE 0, DOES> ( the ((DEFER)) runtime ) 
      @ ?DUP IF EXECUTE THEN ;
 *
 * declare as <c>"DEFER deferword"</c>  <br>
 * and set as <c>"['] executionword TO deferword"</c>
 */
FCode (p4_defer)
{
  p4_header (PFX(p4_defer_RT), 0);
  FX_COMMA (0);
}

/** >EXECUTE ( xt -- ? )
 * same as => EXECUTE , but checks for null as xt argument
 simulate:
   : >EXECUTE  ?DUP IF EXECUTE THEN ;
 */
FCode (p4_to_execute)
{
  p4xt xt = (p4xt) FX_POP;
  
  if (xt) { PFE.execute (xt); }
}

LISTWORDS (lpf83) =
{
  /** the value for BackSpace to be used with => EMIT - compare with => BL */
  OC ("BS",		'\b'),

  CO ("BOUNDS",		p4_bounds),
  CO ("PERFORM",	p4_perform),
  CO ("?LEAVE",		p4_question_leave),
  CO ("NOOP",		p4_noop),
  CO ("RP@",		p4_r_p_fetch),
  CO ("RP!",		p4_r_p_store),
  CO ("SP!",		p4_s_p_store),
  CO ("-ROT",		p4_dash_rot),
  CO ("CSET",		p4_c_set),
  CO ("CRESET",		p4_c_reset),
  CO ("CTOGGLE",	p4_c_toggle),
  CO ("OFF",		p4_off),
  CO ("ON",		p4_on),
  CO ("3DUP",		p4_three_dup),
  CO ("4DUP",		p4_four_dup),
  CO ("UPC",		p4_upc),
  CO ("UPPER",		p4_upper),
  CO ("LOWER",		p4_lower),
  CO ("SKIP",		p4_skip),
  CO ("SCAN",		p4_scan),
  CO ("PLACE",		p4_place),
  CS ("ASCII",		p4_ascii),
  CS ("CONTROL",	p4_control),
  CO ("NUMBER?",	p4_number_question),
  CO ("VOCS",		p4_vocs),
  
  CO ("DEFER",		p4_defer),
  CO (">EXECUTE",	p4_to_execute),
};
COUNTWORDS (lpf83, "FORTH-83-L&P Laxen&Perry extensions");

/*@}*/

