/** @name main.c --- Process command line, get memory and start up.
 * 
 *  Copyright (C) Tektronix, Inc. 1998 - 2000. All rights reserved.
 *
 *  @see     GNU LGPL
 *  @author  Tektronix CTE              @(#) %derived_by: guidod %
 *  @version %version: 4.43 %
 *    (%date_modified: Tue Apr 11 16:59:17 2000 %)
 *
 *  @description
 *  Process command line, get memory and start up the mainloop of PFE
 */
/*@{*/
#if defined(__version_control__) && defined(__GNUC__)
static char* id __attribute__((unused)) = 
      "@(#) $Id: %full_filespec: main.c~4.43:csrc:bln_12xx!13 % $";
#endif

#define _P4_SOURCE 1

#include "p4sys.h"
 
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#ifndef P4_NO_FP
#include <float.h>
#endif
#include <errno.h>
#ifdef HAVE_LOCALE_H
#include <locale.h>
#endif
#include <unistd.h>

#include "support.h"
#include "term.h"
#include "lined.h"
#include "help.h"


#include "nonansi.h"
#include "missing.h"
#include "p4_log.h"

/* only needed static here */
static char* dumpfile = 0;

/************************************************************************/
/* physical instance of the global system variable:                     */
/************************************************************************/

#ifndef P4_REGTH
p4_threadP p4TH;
#endif

/* very simple error handling for fatal errors */

void
p4_fatal (const char *msg,...)
{
  char buf[128] = "";
  va_list p;

  va_start (p, msg);
  vsprintf (buf, msg, p);
  fprintf (stderr, "\n%s: %s (fatal)\n", (char *) P4_opt.argv[0], buf);
  p4_exit (2);
}

void p4_thread_save (p4_threadP p)
{
#ifdef P4_REGIP
  p->ip = p4IP;
#endif
  
#ifdef P4_REGW
  p->wp = p4WP;
#endif
  
#ifdef P4_REGSP
  p->sp = p4SP;
#endif
  
#ifdef P4_REGRP
  p->rp = p4RP;
#endif
  
#ifdef P4_REGLP
  p->lp = p4LP;
#endif

#ifndef P4_NO_FP  
#ifdef P4_REGFP
  p->fp = p4FP;
#endif
#endif
}

void p4_thread_load (p4_threadP p)
{
  p4TH = p; /* either a global variable or a register */

#ifdef P4_REGIP
  p4IP = p->ip;
#endif
  
#ifdef P4_REGW
  p4WP = p->wp;
#endif
  
#ifdef P4_REGSP
  p4SP = p->sp;
#endif
  
#ifdef P4_REGRP
  p4RP = p->rp;
#endif
  
#ifdef P4_REGLP
  p4LP = p->lp;
#endif

#ifndef P4_NO_FP  
#ifdef P4_REGFP
  p4FP = p->fp;
#endif
#endif
}


/************************************************************************/
/* Analyze command line options:                                        */
/************************************************************************/

struct _p4_session p4_def_session =	/* initialized with defaults */
{
  1, (char**)"p4th",		/* argc, argv */
  0,				/* canonical */
  0,				/* stdio */
  CAPS_ON,			/* caps_on */
  LOWER_CASE_ON,		/* lower_case_on */
  LWRCASE_FN_ON,		/* lower_case_fn */
#ifndef P4_NO_FP
  FLOAT_INPUT_ON,		/* float_input */
#else
  0,
#endif
  0,				/* license */
  0,				/* warranty */
  0,				/* quiet */
  0,				/* verbose */
  0,				/* debug */

  TEXT_COLS,			/* cols */
  TEXT_ROWS,			/* rows */

  TOTAL_SIZE,			/* total_size of system in bytes */

#if defined SYS_HPUX && defined CPU_m68k && !defined __GNUC__
  (TOTAL_SIZE / 8) / SIZEOF_CELL,
  (TOTAL_SIZE / 16) / SIZEOF_DOUBLE,
  (TOTAL_SIZE / 16) / SIZEOF_CELL,
#else
  STACK_SIZE			/* stack size in cells */
  ? STACK_SIZE
    : (TOTAL_SIZE / 8) / SIZEOF_CELL,
  FLT_STACK_SIZE		/* floating point stack size in items */
  ? FLT_STACK_SIZE
    : (TOTAL_SIZE / 16) / SIZEOF_DOUBLE,
  RET_STACK_SIZE		/* return stack size in items */
  ? RET_STACK_SIZE
    : (TOTAL_SIZE / 16) / SIZEOF_CELL,
#endif

  MAX_FILES,			/* max_files */
  POCKETS,			/* pockets */
  P4RC_FILE,			/* p4rc_file */
  DEFAULT_BLKFILE,		/* block_file */
  NULL,				/* include_file */
  INC_PATHS,			/* incpath */
  INC_EXTENSIONS,		/* incext */
  BLK_PATHS,			/* blkpath */
  BLK_EXTENSIONS,		/* blkext */
  EDITOR,			/* preferred text file editor */
  P4_MP,			/* initial cpus */
};

static void
howto (p4_sessionP set)
{
  fprintf (stderr,
	   "%s\n%s\n"
	   "Usage: %s [-#bcdefhklrsv] [file]\n"
           "  -1 ... -8" "\tnumber of cpus in this forth [%i]\n"
	   "  -b FILE" "\tuse FILE as block device\n"
	   "  -c\t" "\tturn on CAPS lock [%s]\n"
	   "  -d FILE" "\treload dictionary image from FILE\n"
	   "  -D FILE" "\tbuild dictionary image and exit\n"
	   "  -e NAME" "\tNAME of preferred ASCII text editor [%s]\n"
	   "  -E\t" "\tallow input of floating point numbers [%s]\n"
	   "  -fN\t" "\tmaximum N simultaneously open files [%d]\n"
	   "  -F\t" "\tconvert file names to lower case [%s]\n"
	   "  -h\t" "\tdisplay this message and exit\n"
	   "  -k SIZE" "\tSIZE of system in KBytes [%d K]\n"
	   "  -l\t" "\tallow input of lower case words [%s]\n"
	   "  -L\t" "\tdisplay license\n"
	   "  -p SIZE" "\tSIZE of floating point stack in items [%d]\n"
	   "  -PN" "\tnumber of pockets for S\" [%d]\n"
	   "  -q\t" "\tsuppress signon message\n"
	   "  -r SIZE" "\tSIZE of return stack in cells [%d]\n"
	   "  -s SIZE" "\tSIZE of stack in cells [%d]\n"
	   "  -t CxR" "\ttext screen has C [%d] columns and R [%d] rows\n"
	   "  -v\t" "\tverbose\n"
	   "  -W\t" "\tdisplay warranty. Of course: Absolutely none.\n"
	   "Turn option off by appending \"-\" to the letter.\n"
	   "The given file is loaded initially.\n\n",
	   p4_version_string, p4_copyright_string,
	   set->argv[0],
           set->cpus,
           set->caps_on ? "ON" : "OFF",
           set->editor,
           set->float_input ? "ON" : "OFF",
	   (int) set->max_files,
           set->lower_case_fn ? "ON" : "OFF",
	   (int) set->total_size >> 10,
	   set->lower_case_on ? "ON" : "OFF",
	   (int) set->flt_stack_size,
	   (int) set->pockets,
	   (int) set->ret_stack_size,
	   (int) set->stack_size,
	   (int) set->cols, (int) set->rows);
  p4_exit (1);
}

/* note, that these argc/argv are given as references! */
void
p4_get_options (p4_sessionP set, int argc, char* argv[])
{
  char env_opt[0x400];		/* options from environment variable */
  int i, optc, flag;		/* count of all options */
  char *optv[100];		/* values of all options */
  char opt, *t, *val;
  
  if (!set->cpus)  
     memcpy (set, &p4_def_session, sizeof(struct _p4_session));
     
  set->argv = argv; set->argc = argc;

  /*
   * get special options from environment variables:
   */
  if ((t = getenv ("PFEINCLUDE")) != NULL)
    set->incpaths = strdup (t),
    set->blkpaths = strdup (t);
  if ((t = getenv ("EDITOR")) != NULL)
    set->editor = strdup (t);

  /*
   * merge options from environment variable with those from command line:
   */
  optc = 0;
  if ((t = getenv ("PFEOPTIONS")) != NULL)
    {
      strcpy (env_opt, t);
      t = strtok (env_opt, " ");
      do
	{
	  optv[optc++] = t;
	  t = strtok (NULL, " ");
	}
      while (t);
    }
  set->argc = optc;
  for (i = 1; i < argc; i++)  
    optv[optc++] = argv[i];

  /*
   * process options:
   */
  for (i = 0; i < optc; i++)
    {
      register int l;
      
      t = optv[i];
      if (*t != '-')
	break;
      opt = *++t;
      
      l = strlen(t);
      /*
       * newer flags
       */
      if (l > 6 && !memcmp(t, "-cpus=", 6))
      {  register int cpus = atoi(t+6);
         if (0 < cpus && cpus <= P4_MP_MAX)
            set->cpus = cpus;
         else
            { P4_fail2 ("-cpus=%d invalid (max %d allowed)", cpus, P4_MP_MAX); }
         continue;
      }
      
      if (l > 10 && !memcmp(t, "-dumpfile=", 10))
      {
         dumpfile = t+10;
         continue;
      }
      
      /*
       * Simple flag options can be -x or -x- to turn them off:
       */
      flag = t[1] != '-';
      switch (opt)
	{
	/* *INDENT-OFF* */
	case 'c': set->caps_on = flag;	continue;
	case 'l': set->lower_case_on = flag;continue;
	case 'F': set->lower_case_fn = flag;continue;
	case 'E': set->float_input = flag;	continue;
	case 'L': set->license = flag;	continue;
	case 'W': set->warranty = flag;	continue;
	case 'q': set->quiet = flag;	continue;
	case 'v': set->verbose = flag;	continue;
	case 'B': set->debug = flag;	continue;
	/* *INDENT-ON* */
	}
      /*
       * Other options have values either following immediately after
       * the option letter or as next command line argument:
       */
      if (*++t)
	val = t;
      else if (i == optc - 1)
	val = NULL;
      else
	val = optv[++i];
      switch (opt)
	{
	/* *INDENT-OFF* */
	default:
	case 'h': howto (set);				continue;
	case 'b': set->block_file = val;			continue;
	case 'e': set->editor = val;			continue;
	case 'k': set->total_size = atoi (val) << 10;	continue;
	case 'p': set->flt_stack_size = atoi (val);		continue;
	case 'P': set->pockets = atoi (val);		continue;
	case 'r': set->ret_stack_size = atoi (val);		continue;
	case 's': set->stack_size = atoi (val);		continue;
	/* *INDENT-ON* */
	case 'f':
	  set->max_files = atoi (val);
	  if (set->max_files < 4)
	    set->max_files = 4;
	  continue;
	case 't':
	  if (sscanf (val, "%dx%d", &set->cols, &set->rows) != 2)
	    set->cols = TEXT_COLS, set->rows = TEXT_ROWS;
	  continue;
	case 'I':
	  t = set->incpaths + strlen (set->incpaths);
	  *t++ = PATH_DELIMITER;
	  strcpy (t, val);
	  t = set->blkpaths + strlen (set->incpaths);
	  *t++ = PATH_DELIMITER;
	  strcpy (t, val);
	  continue;
	}
    }
  if (i < optc)
    set->include_file = optv[i++];

  /*
   * Register options starting from included file name in app_ argc/v:
   */
  set->argc = i - set->argc;
  set->argv = &argv[set->argc];
  set->argc = argc - set->argc;
}

/************************************************************************/
/* Initialize memory map:                                               */
/************************************************************************/

static void
init_accept_lined (void)
{
  extern void accept_executes_xt (int);
  static void (*exec[10]) (int) =
  {
    accept_executes_xt, accept_executes_xt, accept_executes_xt,
    accept_executes_xt, accept_executes_xt, accept_executes_xt,
    accept_executes_xt, accept_executes_xt, accept_executes_xt,
    accept_executes_xt,
  };

  memset (&PFE.accept_lined, 0, sizeof PFE.accept_lined);
  PFE.accept_lined.history = PFE.history;
  PFE.accept_lined.history_max = PFE.history_top - PFE.history;
  PFE.accept_lined.complete = p4_complete_dictionary ;
  PFE.accept_lined.executes = exec;
  PFE.accept_lined.caps = P4_opt.caps_on != 0;
}

static void
dict_allocate (void *p, p4ucell size)
/* Allocates all memory areas in a continuous buffer at p with given size. */
{
  void *tmp;
  void *q = (char *) p + size;
  typedef char pock_t[POCKET_SIZE];

#define ALLOC(TYPE,ALIGN,MEMBOT,MEMTOP,SIZE)		\
  (						\
   MEMTOP = (TYPE *) ((size_t)q & ~((size_t)(ALIGN) - 1)), \
   P4_ADD (q, -(SIZE) * sizeof (TYPE)),		\
   MEMBOT = (TYPE *)q			\
  )

  ALLOC (File,	 CELL_ALIGN,	PFE.files,	PFE.files_top, P4_opt.max_files + 3);
  ALLOC (char,	 1,		PFE.history, 	PFE.history_top, HISTORY_SIZE);
  ALLOC (char,	 1,		PFE.tib,	PFE.tib_end, TIB_SIZE);
  ALLOC (p4xt *, CELL_ALIGN,	PFE.rstack,	PFE.r0, P4_opt.ret_stack_size);
  ALLOC (p4cell, CELL_ALIGN,	PFE.stack,	PFE.s0, P4_opt.stack_size);
#ifndef P4_NO_FP
  ALLOC (double, DFLOAT_ALIGN,	PFE.fstack,	PFE.f0, P4_opt.flt_stack_size);
#else
  PFE.fstack = 0; PFE.f0 = 0; 
#endif
  ALLOC (pock_t, 1,		PFE.pockets,	tmp, P4_opt.pockets);
#undef ALLOC

  if ((char *) q < (char *) p + MIN_PAD + MIN_HOLD + 0x4000)
    p4_fatal ("impossible memory map");
  PFE.dict = (p4char *) p;
  PFE.dictlimit = (p4char *) q;
  init_accept_lined ();
}


/************************************************************************/
/* Here's main()                                                        */
/************************************************************************/

/* this is the main multi-forth routine ... */

static void p4_cleanup_mp (void);
extern void p4_mainloop_mp (void);

p4_threadP p4_main_threadP = NULL; /* distinct for each thread !! */

int
p4_main () /* implicit p4TH argument - BEWARE */
{
  auto volatile int N;

#ifdef VXWORKS
  extern int taskVarAdd (int, int*);
  extern int taskIdSelf ();
  taskVarAdd (taskIdSelf (), (int*) &p4_main_threadP);
#endif
  p4_main_threadP = p4TH;  


#ifdef HAVE_LOCALE_H
  setlocale (LC_ALL, "C");
#endif
#if defined SYS_EMX
  _control87 (EM_DENORMAL | EM_INEXACT, MCW_EM);
#endif

  


  for (N=0; N < P4_opt.cpus; N++)
  {
      if (!P4_opt.cpu[N]) 
         continue;

      p4TH = P4_opt.cpu[N];
      switch (setjmp (PFE.loop))
        {
        case 'A':
        case 'Q':
            {
              P4_fatal ("main: Error setting up");
              p4_cleanup_mp ();
              return -1;
            }
        case 'X':
            {
              P4_info ("C ya");
              p4_cleanup_mp ();
              return 0;
            }
        }
  }

  p4TH = p4_main_threadP; /* p4TH = P4_opt.cpu[0]; */

#if !defined __WATCOMC__
  if (!isatty (STDIN_FILENO))
    P4_opt.stdio = 1;
  else
#endif
    {
      P4_opt.stdio = 0;
      if (!p4_prepare_terminal ())
	{
	  if (!P4_opt.quiet)
	    fputs ("[unknown terminal, "
#if defined ASSUME_VT100
		   "assuming vt100"
#elif defined ASSUME_DUMBTERM
		   "assuming dumb terminal"
#else
		   "running in canonical mode"
#endif
		   "]\n", stderr);
#if !defined ASSUME_VT100 && !defined ASSUME_DUMBTERM
	  P4_opt.canonical = 1;
#endif
	}
      p4_interactive_terminal ();
      PFE.system_terminal = &p4_system_terminal;
    }

  p4_install_signal_handlers ();
    
  if (!P4_opt.quiet)
    {
      p4_outs ("\\ ");
      p4_outs (p4_version_string);
      p4_outs (p4_copyright_string);
      if (P4_opt.license)
	p4_outs (p4_license_string);
      if (P4_opt.warranty)
	p4_outs (p4_warranty_string);

/* BYE doesn't make sense in an embedded system. */
#ifndef _K12_SOURCE
      p4_outs ("\n\nTo quit say BYE.");
#endif /* _K12_SOURCE */
      if (!P4_opt.license || !P4_opt.warranty)
	p4_outs ("\n\nPlease enter LICENSE and WARRANTY.\n");
      else
        p4_outs ("\n\nHi there, enjoy Forth!\n");
    }
  if (PFE.rows == 0)
    PFE.rows = P4_opt.rows;
  if (PFE.cols == 0)
    PFE.cols = P4_opt.cols;

  for (N=0; N < P4_opt.cpus; N++)
  {

    if (!P4_opt.cpu[N])
        continue;
    else
        p4TH = P4_opt.cpu[N];

    p4TH->system_terminal = P4_opt.cpu[0]->system_terminal;
    p4TH->cleanup = &p4_cleanup_mp;

    PFE.memory = p4_xcalloc (1, (size_t) P4_opt.total_size);
    dict_allocate ((char *) PFE.memory, P4_opt.total_size);

    P4_info3 ("forth%i new mem at %p len %d", N, PFE.memory, P4_opt.total_size);

    p4_initialize_system ();
  }

  p4TH = p4_main_threadP; /* p4TH = P4_opt.cpu[0]; */

  /* If running in a pipe, process commands from stdin: */
  if (P4_opt.stdio)
    {
      p4_include_file (PFE.stdIn);
      PFE.cleanup ();
      return 0;
    }

  /* If it's a turnkey-application, start it: */
  if (APPLICATION)
    {
      p4_run_forth (APPLICATION);
      PFE.cleanup ();
      return 0;
    }
  if (P4_opt.verbose)
    FX (p4_dot_memory);
    
  if (dumpfile)
  {  int f = open(dumpfile, O_WRONLY|O_CREAT, 0660);
     if (-1!=f) { write (f, PFE.dict, PFE.dp-PFE.dict); close (f); }
  }

  p4_mainloop_mp (); /* p4_main is using multi-forth-mainloop, uhh */
  PFE.cleanup ();
  return 0;
}

#if 0
static void
p4_cleanup ()
{
  extern void p4_forget_on_exit ();
  p4_forget_on_exit ();

  if (PFE.system_terminal) 
     PFE.system_terminal ();
  if (PFE.memory)			/* non-zero only, if calloced */
    {
      p4_xfree (PFE.memory);
      PFE.memory = 0;
    }
}
#endif

static void
p4_cleanup_mp ()
{
  extern void p4_forget_on_exit ();
  extern void p4_cleanup_terminal ();
  int n;
  
  P4_enter ("cleanup forth");

  if (P4_opt.cpu[0])
     p4TH = P4_opt.cpu[0];  /* cpu[0] should be always valid */

  if (PFE.system_terminal)    /* call this once, with the first cpu */
     PFE.system_terminal ();
     
  p4_cleanup_terminal ();

  for (n=P4_opt.cpus; n >= 0; n--)
  {
      if (!P4_opt.cpu[n]         /* no such cpu allocated */
      ||  !P4_opt.cpu[n]->dict ) /* or this cpu is just void */
         continue;
         
      P4_debug1 (1, "cleanup forth%i", n);

      p4TH =P4_opt.cpu[n];
      
      p4_forget_on_exit ();
      
      { /* see if there's some memory chunk still to be freed */
         register void** p = &PFE.p[0];
         for ( ; p <= &PFE.memory; p++) {
            if (*p) { 
               P4_info2 ("forth%i free %p", n, *p);
               p4_xfree (*p); 
               *p = 0; 
            }
         }
      }
  }

  P4_leave ("cleanup done");
}

#ifdef STATIC_MAIN
static int
static_main (int argc,
             char** argv)
#else /* STATIC_MAIN */
int
main (int argc,
      char** argv)
#endif /* STATIC_MAIN */
{
  struct _p4_session session;
  int i;
  
  memcpy (&session, &p4_def_session, sizeof(struct _p4_session));
  p4_get_options (&session, argc, argv);
  if (session.cpus < 1) session.cpus = 1;
  if (session.cpus > P4_MP_MAX) session.cpus = P4_MP_MAX;

  p4TH = malloc (sizeof(struct _p4_thread)*session.cpus);
  if (!p4TH) return -ENOMEM;
  memset (p4TH, 0, sizeof(struct _p4_thread)*session.cpus);
  
  for (i=0; i<session.cpus; i++)
  {
     session.cpu[i] = &p4TH[i];
     p4TH[i].set = &session;
     p4TH[i].nr = i;
  }
  return p4_main (); /* implicit p4TH-argument ! */
}

#ifdef STATIC_MAIN
void
p4e (int arg1, int arg2, int arg3, int arg4, int arg5,
     int arg6, int arg7, int arg8, int arg9, int arg10)
{
  int argc = 0;
  char *argv[12];

  argv [0] = "p4th";
  argv [1] = (char *) arg1;
  argv [2] = (char *) arg2;
  argv [3] = (char *) arg3;
  argv [4] = (char *) arg4;
  argv [5] = (char *) arg5;
  argv [6] = (char *) arg6;
  argv [7] = (char *) arg7;
  argv [8] = (char *) arg8;
  argv [9] = (char *) arg9;
  argv[10] = (char *) arg10;
  argv[11] = NULL;
  while (argv[argc]!=NULL)	/* count args */
  {
    argc++;
  }
  static_main (argc, argv);
}
#endif 
   /* STATIC_MAIN */
/*@}*/

