/** @name PFE-SYSTEM --- Words making sense in POSIX-like systems only.
 *
 *  Copyright (C) Tektronix, Inc. 1998 - 2000. All rights reserved.
 *
 *  @see     GNU LGPL
 *  @author  Tektronix CTE                %derived_by: guidod %
 *  @version %version: 4.54 %
 *    (%date_modified: Fri May 05 19:02:37 2000 %)
 *
 *  @description
 *		This file exports a set of system words for 
 *              a posixish OS environment. So should do
 *		any alternative wordset you might create for your OS.
 */
/*@{*/
#if defined(__version_control__) && defined(__GNUC__)
static char* id __attribute__((unused)) = 
      "@(#) $Id: %full_filespec: unix.c~4.54:csrc:bln_12xx!4 % $";
#endif

#define _P4_SOURCE 1

#include "p4sys.h"

#include <stdio.h>
#include <errno.h>
#include <limits.h>
#include <string.h>
#include <time.h>

#include "support.h"

/** 
 * CLOCKS_PER_SEC - usually one million ticks, but can be
 * <i>very</i> different on a specific system. Exported
 * as a constant. see => CLOCK
 */  /*"CLK_TCK"*/
#ifdef CLOCKS_PER_SEC
#  undef CLOCKS_PER_SEC
#endif
#define CLOCKS_PER_SEC 1000000

#include "nonansi.h"
#include "missing.h"
#include "p4_log.h"

extern Head *
   p4_make_head (const char *name, int count, char **nfa, Wordl * wid);
   
#ifndef PATH_MAX
# ifdef _POSIX_PATH_MAX
# define PATH_MAX _POSIX_PATH_MAX
# else
# define PATH_MAX 255
# endif
#endif


/* .......................................
  system specific module loader handlers
*/

#ifdef HAVE_DLFCN_H
   /* esp. Linux and Solaris */
#define HAVE_P4_DLFCN

#include <dlfcn.h>

/* dl-specific: pointer to program symbol table */
static void* p4_dlself = 0;

/* dl-specific: init dl symbol table, dl error */
int 
p4_dlinit ()
{
  if (!p4_dlself )  
     p4_dlself = dlopen (0, RTLD_NOW|RTLD_GLOBAL);
     
  if (!p4_dlself )
     return -ENOEXEC;
  else
     return 0;
}

/* dl-specific: describe last dl-error (stderr) */
void 
p4_dlerror (const char* where)
{
  fprintf (stderr, "%s: %s", where, dlerror ());
}  

/* dl-specific: load shared-object into program codespace */
void* 
p4_dlopen (const char* dir, const char* name)
{
  void* lib;
  char libname[PATH_MAX+1];
  
  if (!name) return 0;
  if (!p4_dlself ) p4_dlinit ();

  if (dir && dir[0]) 
  { 
     int len = strlen (dir);
     if (len > PATH_MAX-2) return NULL;
     memcpy (libname, dir, len+1);
     if (libname[len-1] != '/') strcat (libname+len-1, "/");
  } else {
     *libname = 0;
  }
  
  strncat (libname, name, PATH_MAX);
  strncat (libname, ".so", PATH_MAX);
  
  lib = dlopen (libname, RTLD_NOW|RTLD_GLOBAL);
  
  return ((void*) lib);
}

/* dl-specific: remove shared-object from program codespace */
int 
p4_dlclose (const void* lib)
{
  return dlclose ((void*)lib);
}

/* dl-specific: find symbol in loaded object */
void* 
p4_dlsym (const void* lib, const char* symbol)
{
  if (!symbol) return 0;

  if (!lib)
     return dlsym ((void*)p4_dlself , symbol);
  else
     return dlsym ((void*)lib, symbol);
}

/* --- */

#elif defined SYS_hpux
#define HAVE_P4_DLFCN

#include <dl.h>

/* hp-specific: pointer to program symbol table */
static void* p4_dlself = 0;

/* hp-specific: init dl symbol table, dl error */
int 
p4_dlinit ()
{
  p4_dlself = (void*)PROG_HANDLE;
  return 0;
}

/* hp-specific: describe last dl-error (stderr) */
void 
p4_dlerror (const char* where)
{
  perror (where);
}  

/* hp-specific: load shared-object into program codespace */
void* 
p4_dlopen (const char* dir, const char* name)
{
  void* lib;
  char libname[PATH_MAX+1];
  
  if (!name) return NULL;
  if (!p4_dlself ) p4_dlinit ();

  if (dir && dir[0]) 
  { 
     int len = strlen (dir);
     if (len > PATH_MAX-2) return NULL;
     memcpy (libname, dir, len+1);
     if (libname[len-1] != '/') strcat (libname+len-1, "/");
  } else {
     *libname = 0;
  }
  
  strncat (libname, name, PATH_MAX);
  strncat (libname, ".sl", PATH_MAX);
  
  lib = shl_load (libname, BIND_IMMEDIATE|BIND_NONFATAL, 0);
  
  return ((void*) lib);
}

/* hp-specific: remove shared-object from program codespace */
int 
p4_dlclose (const void* lib)
{
  return shl_unload ((shl_t)lib);
}

/* hp-specific: find symbol in loaded object */
void* 
p4_dlsym (const void* lib, const char* symbol)
{
  shl_t my_lib;
  void* val;

  if (!symbol) return NULL;
  
  if (!lib) 
      my_lib = (shl_t)p4_dlself ;
  else
      my_lib = (shl_t)lib;
      
  if (shl_findsym (&my_lib, symbol, 0, &val) == (-1)) 
      val = NULL;
  
  return val;
}

#elif defined SYS_vxworks || defined VxWorks
#define HAVE_P4_DLFCN

#include <symLib.h>
#include <loadLib.h>
#include <unldLib.h>

/* vx-specific: pointer to program symbol table */
static void* p4_dlself = 0;

/* vx-specific: init dl symbol table and dl error */
int 
p4_dlinit ()
{
  extern SYMTAB_ID sysSymTbl; /* the global SymTbl of loadModule */
  
  p4_dlself = (void*)sysSymTbl;

  return 0;
}

/* vx-specific: describe last dl-error -> divert to p4log */
void 
p4_dlerror (const char* where)
{
  P4_fail2 ( "%s: (errno=%s)", where, strerror(errno));
}  

/* vx-specific: load shared-object into program codespace */
void* 
p4_dlopen (const char* dir, const char* name)
{
  int fd;
  MODULE_ID lib = 0;
  char libname[PATH_MAX+1];
  
  if (!name) return NULL;
  if (!p4_dlself ) p4_dlinit ();

  if (dir && dir[0]) 
  { 
     int len = strlen (dir);
     if (len > PATH_MAX-2) return NULL;
     memcpy (libname, dir, len+1);
     if (libname[len-1] != '/') strcat (libname+len-1, "/");
  } else {
     *libname = 0;
  }
  
  strncat (libname, name, PATH_MAX);
  strncat (libname, ".O", PATH_MAX);

  fd = open (libname, O_RDONLY, 0);
  if (fd != ERROR) {
    P4_enter1("ld < '%s'", libname);
    lib = loadModule (fd, GLOBAL_SYMBOLS); /* or ALL_SYMBOLS *askmee*/
    P4_leave1("ld done = %p", lib);
    close (fd);
  }
  
  return ((void*) lib);
}

/* vx-specific: remove shared-object from program codespace */
int 
p4_dlclose (const void* lib)
{
  return ((int)
     (unldByModuleId ((MODULE_ID)lib, 0)));
}

/* vx-specific: find symbol in loaded object -
   BEWARE: vxworks uses the same symbol table for all shared-objects,
   so you may find a symbol from another lib - the argument is ignored! 
*/
void *
p4_dlsym (const void* lib, const char* symbol)
{
  void* val;
  lib=lib; /* lib is ignored - suppress compiler warning */
  
  if (!symbol) return 0;
  
  if (symFindByName ((SYMTAB_ID) p4_dlself, (char*) symbol, (char**) &val, 0)
  == ERROR)
  {
     if (strlen(symbol) < 127) 
     {
        auto char _symbol[128];
        _symbol[0] = '_';  strcpy (&_symbol[1], symbol);

        if (symFindByName ((SYMTAB_ID) p4_dlself, _symbol, (char**) &val, 0)
        == ERROR)
        {
          val = 0;
        }
     }
  }

  return val;
}

#endif /* HAVE_P4_DLFCN layer */

/* the define HAVE_P4_DLFCN notifies the rest of the file
   if we have a p4_dlfcn layer of dynamic loading of objects
   ...................................................
*/


#ifdef HAVE_USELIB
FCode (p4_uselibrary)
{
  *SP = uselib ((char *) *SP);
}
#endif

#if defined SYS_linux && !defined (__cplusplus)
/* ********************************************************************** */
/* Linux shared library calls -- KAH 930824				  */
/* ********************************************************************** */

/* direct call to shared library function (without loadlist)   */

typedef struct
  {
    unsigned nargs	:4;	/* # of arguments */
    unsigned restype	:2;	/* result type */
    unsigned		:2;	/* reserved -- complex res */
    unsigned argtype	:2;	/* double/long arg flags */
    unsigned		:22;	/* other argtypes shifted */
				/* from here */
  }
control_word;

#define DYN_INTEGER	0
#define DYN_LONGINT	1
#define DYN_FLOAT	2
#define DYN_LONGFLOAT	3

#define _exec(sub,resulthi,resultlo)	\
	__asm__ __volatile__ ("call *%2;movl %%edx, %0;movl %%eax, %1":	\
	"=g" (resulthi), "=g" (resultlo):"g" (sub): "eax", "edx");
#define _cpush(x)	\
	__asm__ __volatile__ ("pushl %0;"::"g" (x));

/* not sure if this float stuff is right.  Are singles and 
   doubles same length?  Are these stored in a different forth stack?
   How do I copy things back and forth from int/pointer stack to
   float stack? */

#ifndef P4_NO_FP
#define lowfresult(x)	__asm__ __volatile__ ("fstpl %0;":"=g" (x));
#define highfresult(x)	lowfresult(x);
#endif

static void
call_c (p4code * sub)
{
  int i, high, low, result_type;
  p4udcell *tmp;
  control_word x = *(control_word *) SP++;

  i = x.nargs;
  result_type = x.restype;
#ifndef P4_NO_FP
  tmp = (p4udcell *) FP;
#else
  tmp = (p4udcell *) SP;
#endif
  while (i--)
    {
      switch (x.argtype)
	{
	case DYN_LONGINT:
	  _cpush (*SP++);
	case DYN_INTEGER:
	  _cpush (*SP++);
	  break;
#ifndef P4_NO_FP
	case DYN_LONGFLOAT:
	  _cpush (tmp->hi);
	case DYN_FLOAT:
	  _cpush (tmp++->lo);
#endif
	}
      *(p4ucell *) & x >>= 2;
    }
  _exec (sub, high, low);
  switch (result_type)
    {
    case DYN_INTEGER:
      *--SP = low;
      break;
    case DYN_LONGINT:
      *--SP = low;
      *--SP = high;
      break;
#ifndef P4_NO_FP
    case DYN_FLOAT:
      lowfresult (*--tmp);
      break;
    case DYN_LONGFLOAT:
      lowfresult (*--tmp);
      highfresult (*--tmp);
#endif
    }
}

FCode (p4_call_c)
{
  call_c ((p4code *) * SP++);
}

#endif /* Linux shared library calls */


/* ********************************************************************* */

/*
   The following section defines the funtions necessary to support
   forth's dynamically loading modules. It is split into
   two parts: one is loading shared-objects into the task's
   code-space, the other assigns slots in the threaded forth-structure
   that are used to contain a pointer to an object-local data-space.
   
   We will assume, that a single threaded register-based pointer p4TH
   exists which points to a structure holding pointers to
   dl-local mem-space. The id's for the data-space must 
   be given per code-space, so unlike most other data-spaces, 
   here we have them made static here - and it ain't no error.
   
   The slot-id can be both load-dynamic or compile-static.
   In dynamic mode, the code-space slot-var is null before, and
   after slot_use it contains the slot-index grabbed from the
   pool of free slot-indices.
   In compiled mode, the code-space slot-var contains a slot-index;
   slot_use accepts this as a request to a specific slot, and
   does only check if it had still been free, otherwise fails
   without assigning another number from the pool of free slots.
   
   All forth-machines using the same shared-object which in turn
   needs a threaded data-space, will have the same slot-index
   in use.
   
   slot 0 is valid but not assignable - it's the forth dictionary space.
   It cannot be touched by any of the following calls. The
   slot 0 is only touched by the cleanup-routines that calls
   mem-free for all alloc'ed slots in a thread.
*/

#undef  INBETWEEN
#define INBETWEEN(x,a,e) ((a) < (x) && (x) < (e))

#undef  BETWEEN
#define BETWEEN(x,a,e) ((a) <= (x) && (x) < (e))

struct slot_info
{
  int* var;          /*(self-referential) ptr to global var holding slot id*/
  short use;         /* use count */
  char  got;         /* (*var) had been null before we had written to it */
};

static struct slot_info p4_slot_table[P4_MOPTRS] = {{0}}; 

/** 
 * request a slot index. The index is written to the variable arg-address.
 * if the arg-address contains a value != 0, we check if that specific
 * slot index is free for assignment - or already assigned to this variable.
 * a slot_use can be done for the same slot-variable multiple times,
 * which will increase a use-counter. Call slot_unuse correspondingly.
 */
int
p4_slot_use (int* var)
{
  if (!var || !BETWEEN(*var,0,P4_MOPTRS)) 
     return -EINVAL;

  if (*var)  /* has an id */
  {
     if (var != p4_slot_table[*var].var) 
     {
        if (!p4_slot_table[*var].var)  /* not used so far */
        {
           p4_slot_table[*var] .var = var;
           p4_slot_table[*var] .use = 1;
           return 0;
        } else {
           return -EACCES;
        }
     } else { /* reuse id */
        p4_slot_table[*var].use++;
        return 0;
     }
  } else {
     /* new id */
     register int i;
     for (i = 1; i < P4_MOPTRS; i++) 
     {
        if (!p4_slot_table[i].var) 
        {
           p4_slot_table[i] .var = var;
           p4_slot_table[i] .got = 1; /* we're writing to var now */
           p4_slot_table[i] .use = 1;
           *var = i;
           return 0;
        }
     }
     return -EBUSY;
  }
}

/**
 * => p4_slot_use
 */
int
p4_slot_unuse (int* var)
{
  if (!var || !INBETWEEN(*var,0,P4_MOPTRS)
  || p4_slot_table [*var].var != var)
     return -ENOENT;
     
  if (! --p4_slot_table [*var].use) 
  {   /* real free of slot */
      register int slot = *var;

      if (p4_slot_table[slot].got) /* had written to var */
         *var = 0;                 /* so be careful and reset it */
         
      memset (&p4_slot_table[slot], 0, sizeof (struct slot_info));
  } 
  
  return 0;
}  


/* ============= and here's the loading part ================= */

/* here again, dlslot 0 is invalid for dynamic loading. It's not
   actually reserved, but, well handy to write
*/

#define P4_DLSLOTS P4_MOPTRS

/* keep `name` to be the first entry, so we can cast to (const char*) */
struct dlslot
{
  char name[256];   /* cannot be longer than longest forth-string */
  void* dlptr;      /* the dlopen handle */
  void* (*llist)(); /* loadlist extraction */
  int use;          /* usecount */
};

typedef void* (*p4_llist_f)(); /* loadlist extraction typedef */

struct dlslot p4_dlslot_table [P4_DLSLOTS] = {{ "", 0 }};


/* copies a forth-module-name into
   name-area being a C-string afterwards.
   A possible shared-object extension is gotten stripped away 
 */
static char* 
module_makename (char* name, const char* nameptr, int namelen)
{
   memcpy (name, nameptr, namelen);
   name [namelen] = 0;

   if (namelen > 2 
   && ( !memcmp (&name[namelen-2], ".o", 2)
     || !memcmp (&name[namelen-2], ".O", 2)
      ))
   {
     name [namelen-2] = 0; /* truncate to basename  */
   }else
   if (namelen > 3 
   && ( !memcmp (&name[namelen-3], ".so", 3)
     || !memcmp (&name[namelen-3], ".sl", 3)
      ))
   {
     name [namelen-3] = 0; /* truncate to basename */
   }
   
   return name;
}

static int
p4_dlslot_find (const char* nameptr, int namelen)
{
  int slot;
  int len;
  char name[256];
  
  module_makename (name, nameptr, namelen);
  len = strlen (name);
  
  for (slot = 1; slot < P4_DLSLOTS; slot++)
  {
    if (!memcmp (p4_dlslot_table[slot].name, name, len+1))
       return slot; /* found */
  }
  
  return 0;
}

static int
p4_dlslot_create (const char* nameptr, int namelen)
{
  int slot;
  char len;
  char name[PATH_MAX+1];
  
  module_makename (name, nameptr, namelen);
  len = strlen (name);
  
  for (slot = 1; slot < P4_DLSLOTS; slot++)
  {
    if (!memcmp (p4_dlslot_table[slot].name, name, len+1)) {
       return slot; /* found */
    }
  }
  /* none found, need new slot */
  
  for (slot = 1; slot < P4_DLSLOTS; slot++) {
     if (!p4_dlslot_table[slot].use) {
        p4_dlslot_table[slot].use = 1;
        memcpy (p4_dlslot_table[slot].name, name, len+1);
        p4_dlslot_table[slot].dlptr = 0; /* mark just created */
        return slot; /* new */
     }
  }
  
  return -EBUSY; /* none found, table full, we're very sorry, dave */
}

static void
p4_dlslot_remove (int slot)
{
  if (!INBETWEEN(slot, 0, P4_DLSLOTS))
  {
     P4_warn1 ("dlslot %i out of range", slot);
     return;
  }
  
  memset (&p4_dlslot_table[slot], 0, sizeof (struct dlslot));
}  


static int
p4_dlslot_isnt_unique_llist (int slot)
{
  int i;

  if (!INBETWEEN(slot, 0, P4_DLSLOTS))
     return -EINVAL;
     
  if (!p4_dlslot_table[slot].use
  ||  !p4_dlslot_table[slot].llist)
     return -ENOENT;

  for (i = 1; i < P4_DLSLOTS; i++)
  {
    if (i == slot)
       continue;
       
    if (p4_dlslot_table[slot].llist && p4_dlslot_table[i].llist 
    &&  p4_dlslot_table[slot].llist == p4_dlslot_table[i].llist)
       return 1; /* isnt unique */
  }

  return 0; /* is unique */
}

#ifdef HAVE_P4_DLFCN

#ifndef MODULE_DIR
#  ifdef P4_MODULE_DIR
#    define MODULE_DIR P4_MODULE_DIR
#  else
#    define MODULE_DIR LIBDIR DIR_DELIMSTR PACKAGE
#  endif
#endif

/* negative return values denote error conditions */
int
p4_dlslot_open (const char* nameptr, int namelen)
{
  int slot;
  void* dll;

  if ((slot= p4_dlslot_find (nameptr, namelen)))
  {
     p4_dlslot_table[slot].use++;
     P4_info2 ("module already loaded: [%i] '%s'", slot, p4_dlslot_table[slot].name);

     return slot;
  } /*else*/

  slot= p4_dlslot_create (nameptr, namelen);   

  P4_enter1 ("loading '%s'", p4_dlslot_table[slot].name);
  
  dll = p4_dlopen ("",p4_dlslot_table[slot].name);
  if (!dll)
     dll = p4_dlopen (MODULE_DIR, p4_dlslot_table[slot].name);

  if (!dll) 
  { 
     p4_dlerror (p4_dlslot_table[slot].name);
     p4_dlslot_remove (slot);
     return -ENOENT; 
  }

  p4_dlslot_table[slot].dlptr = dll;  
  
  p4_dlslot_table[slot].llist = p4_dlsym (dll, "p4_module_llist");
  if (p4_dlslot_isnt_unique_llist (slot)) 
      p4_dlslot_table[slot].llist = NULL;
  
  if (!p4_dlslot_table[slot].llist) 
  {
     P4_fail1 ("'%s': cannot find modules loadlist", p4_dlslot_table[slot].name);
     p4_dlclose (dll);
     p4_dlslot_remove (slot);
     return -ESRCH;
  }
  
  P4_leave1 (" dlslot = %i", slot);
  return slot;
}

void
p4_dlslot_close (int slot)
{
  if (!INBETWEEN(slot,0,P4_DLSLOTS)) {
     P4_warn1 ("dlslot %i out of range", slot);
     return;
  }
     
  if (! --p4_dlslot_table[slot].use) 
  {
      if (p4_dlclose (p4_dlslot_table[slot].dlptr)) 
          p4_dlerror ("p4_dlslot_close"); 
      else
          P4_note1 ("done dlunmap: '%s'", p4_dlslot_table[slot].name); 
  
      p4_dlslot_remove (slot);
  } else {
      P4_leave1 ("kept dlmap: '%s'", p4_dlslot_table[slot].name);
  }
}

void
p4_forget_loadm (void)
{
  int slot = WP_PFA[0];
  
  P4_note1 ("unuse dlslot %i", slot);
  
  p4_dlslot_close (slot);
}

/** (LOADM) ( bstring -- sucess )
 */   
FCode (p4_loadm2)
{
  int slot; 
  
  P4_enter2("dlmap of '%.*s'", **(char**)SP, *(char**)SP+1); 
  
  slot = p4_dlslot_open ((*(char**)SP)+1, **(char**)SP); 

  P4_info1 ( "dlslot: %d", slot);
  if (slot < 0) { *SP = 0; return; }
  
  p4_forget_word ("%s.O", (p4cell) p4_dlslot_table[slot].name, 
                	 p4_forget_loadm, (p4cell) slot);

  if (p4_dlslot_table[slot].llist) 
  {
    void* llist = (*p4_dlslot_table[slot].llist)();

    p4_load_llist (llist, ONLY, 1); /* load the words */
    
    *SP = (p4cell)( llist );
  } else {
    *SP = 0;
  }

  P4_leave1 ("done loadlist at 0x%x", *SP);
}

/** LOADM filename ( -- )
 *  dlmap the shared object (or share an already mapped object)
 *  and run the per-thread initialization code. This is the
 *  user-convenient function, otherwise use => (LOADM)
 simulate:
   : LOADM  BL WORD   
     ((IS_MODULE_LOADED)) IF EXIT THEN 
     HERE (LOADM)  0= IF ." -- load failed: " HERE COUNT TYPE CR THEN ;
 */
FCode (p4_loadm)
{
   char* fn = p4_word (' ');
   
   { 
     char name[260]; module_makename (name, fn+1, *fn);
     strcat (name, ".O"); /* ... name *is* long enough ... */
     if (p4_search_wordlist (name, strlen(name), PFE.atexit_wl))
     { P4_info1 ("%s: is loaded (skipped)", name);
       return; /* already loaded */
     }
   }

   FX_PUSH (fn);
   FX (p4_loadm2);
   if (FX_POP) return;
   p4_outf (" -- load failed: '%.*s'\n", *fn, fn+1);
}

/** DLSYM symbolname ( -- address ) exec-only
 *  lookup the symbol that follows and leave the address (or null)
 */
FCode (p4_dlsym)
{
  char* p; int plen;
  char name[256];
  
  FX (p4_Q_exec); /* currently ignored while compiling */
  
  p4_parseword (' ', &p, &plen);
  
  if (plen) {
     memcpy (name, p, plen);
     name[plen] = '\0';
     FX_PUSH (p4_dlsym (0, name));
     
     if (! *SP) {
        P4_warn1 ("no dlsym '%s'", name);
     }
  }
}

/* DLCALL symbolname ( x8 ... x2 x1 -- x8 ... x2 x1 ) exec-only
 *  does not do any stack changes, the caller
 *  must clean up the stack himself. You don't even
 *  get the return value (sigh)
*/
FCode (p4_dlcall)
{
  char* p; int plen;
  char name[256];
  void (*f)();
  
  FX (p4_Q_exec); /* currently ignored while compiling */
  
  p4_parseword (' ', &p, &plen);
  
  if (plen) {
     memcpy (name, p, plen);
     name[plen] = '\0';
     f = p4_dlsym (0, name);
     
     if (f) { 
        (*f)(SP[0], SP[1], SP[2], SP[3], SP[4], SP[5], SP[6], SP[7]);
     }
     else {
        P4_warn1 ("no dlsym '%s'", name);
     }
  }
}

#endif
   /* HAVE_P4_DLFCN */

#if !(defined SYS_EMX || defined SYS_WATCOM)

#define DEFINED_ignore_line
/** #! ... ( -- ) 
 * ignores the rest of the line,
 * defining `#!' is used to support forth scripts 
 * executed by the unix kernel
 */
FCode (p4_ignore_line)
{
  p4_refill ();
}

#endif

/** CLOCK ( --- ticks )	
 * return clock() 
 */
static
FCode (clock)			
{
  *--SP = (p4cell)clock ();
}

LISTWORDS (system) =
{
#if defined SYS_linux && !defined __cplusplus
  /** accessing shared libraries - only available for i386-linux */
  CO ("CALL-C",		p4_call_c),
#endif
#ifdef HAVE_USELIB
  CO ("USELIBRARY",	p4_uselibrary),
#endif

#if defined HAVE_P4_DLFCN
  CO ("(LOADM)",        p4_loadm2),
  CO ("LOADM",    	p4_loadm),
  
  CI ("DLSYM",          p4_dlsym),
  CI ("DLCALL",         p4_dlcall),
#endif

#ifdef DEFINED_ignore_line
  CO ("#!",		p4_ignore_line),
#endif
  CO ("CLOCK",		clock),
#if defined SYS_AIX1
  OC ("CLK_TCK",	1000000) /**/,        
#elif defined CLOCKS_PER_SEC
  OC ("CLK_TCK",	CLOCKS_PER_SEC) /**/, 
#elif defined CLK_TCK
  OC ("CLK_TCK",	CLK_TCK), 
#else
  OC ("CLK_TCK",	1000000) /**/, /* just a guess :-) */
#endif
};
COUNTWORDS (system, "PFE-SYSTEM system words - mostly posix");

/*@}*/

