/** @name PFE-SMART --- additional primitives
 *
 *  Copyright (C) Tektronix, Inc. 1998 - 2000. All rights reserved.
 *
 *  @see     GNU LGPL
 *  @author  Tektronix CTE              @(#) %derived_by: guidod %
 *  @version %version: 4.49 %
 *    (%date_modified: Mon May 15 17:54:43 2000 %)
 *
 *  @description
 *              This wordset is the place to add any additional primitives
 *		you might wish. A set of words do already live here that
 *              must be bound statically into the main pfe-object to
 *              work out smart and nicely.
 */
/*@{*/
#if defined(__version_control__) && defined(__GNUC__)
static char* id __attribute__((unused)) = 
      "@(#) $Id: %full_filespec: yours.c~4.49:csrc:bln_12xx!4 % $";
#endif
/**
 *
 * To make a new primitive Forth word, you have to write a C function of
 * type
 *
 *	static void name_(void)
 *
 * Use the preprocessor macro
 *
 *	FCode (p4_name)
 *
 * to provide the prototype with the underscore appended to the
 * name. The underscore helps to avoid name clashes with names and
 * keywords of the C language.
 *
 * If you want to call such a primitive from C-source you must of
 * course append the underscore to the name in the function call
 * yourself. If you want to call it from another input file then you
 * must leave out the `static' keyword in the function definition,
 * which is what the alternative macro
 *
 *	code (name)
 *
 * does. Both are of course nothing but shortcuts to ease input of all
 * those words.
 *
 * Having defined a primitive you must add it to the forth dictionary
 * in order to make it visible to Forth.  At the end of each input
 * file defining forth primitives you'll find a table declared with
 * the macro
 *
 *	LISTWORDS (wordset_name) =
 *	{
 *	  ...
 *	};
 *
 * This declares a vector of some type and shows how to continue with
 * the initialization of the vector.  Each vector element registers a
 * forth word for loading into the dictionary at startup. In p4th the
 * dictionary is not an initialized C structure but it is built from
 * such structures -- like the one described here -- at startup.
 *
 * To enter words into this table there are several macros, all for
 * brevity named with two uppercase letters. Using such a macro adds
 * one more element to the list of initializers following the
 * LISTWORDS declaration. Besides primitives you can enter variables
 * and constants to the dictionary this way.
 *
 *	CO (NAME, c-name)	a primitive, not immediate
 *	CI (NAME, c-name)	an immediate primitive
 *	CS (NAME, c-name)	a primitive with separate execution
 *				and compilation semantics, see examples
 *				in core.c (if, case etc.) and macros in
 *				p4_comp.h and -- good luck :-)
 *	OV (NAME)		a normal variable
 *	OC (NAME, value)	a normal constant
 *	OL (NAME, value)	a normal value
 *	IV (NAME)		an immediate variable
 *	IC (NAME, value)	an immediate constant
 *	IL (NAME, value)	an immediate value
 */
 
#define _P4_SOURCE 1
#include "p4sys.h"

#include <string.h>
#include <ctype.h>
#include "support.h"
#include "p4_log.h"

extern Head *
   p4_make_head (const char *name, int count, char **nfa, Wordl * wid);


/** (IMMEDIATE#) ( -- bit-mask )
 *  returns the bit-mask to check if a found word is immediate
    " my-word" FIND IF >FFA C@ (IMMEDIATE#) AND 
                       IF ." immediate" THEN ELSE DROP THEN
 */
FCode (p4_immediate_bit)
{
  FX_PUSH (P4xIMMEDIATE);
}

/** (SMUDGE#) ( -- bit-mask )
 *  returns the bit-mask to check if a found word is smudge
    " my-word" FIND IF >FFA C@ (SMUDGE#) AND 
                       IF ." smudge" THEN ELSE DROP THEN
 */
FCode (p4_smudge_bit)
{
  FX_PUSH (P4xSMUDGED);
}

/** 
 * see => >COMPILE  and => POSTPONE
 */
void p4_to_compile (p4xt xt)
{
  if (!xt) return;
  if (STATE && !(*_FFA(p4_to_name (xt)) & P4xIMMEDIATE))
     COMMA (xt);
  else
     p4_call (xt);
}

/** >COMPILE ( xt -- )
 *  does the work of => POSTPONE on the execution token that 
 *  you got from somewhere else - so it checks if the name
 *  (that correspond to the execution-token argument) is
 *  actually immediate, so it has to be executed to compile
 *  something, e.g. => IF or => THEN - see also => POSTPONE ,
 *  => COMPILE , => [COMPILE] , => INTERPRET
 */
FCode (p4_to_compile)
{
  p4_to_compile ((p4xt) FX_POP);
}

#define PAREN_MAGIC MAKE_MAGIC('P','(',')','X')

extern FCode (p4_tick);
/** ($ word  ( -- cs-token ) compile-only
 *  takes the execution token of the following word and
 *  saves it on the compile-stack. The correspondig closing
 *  => ) will then feed it into => >COMPILE - so this pair
 *  of word provides you with a prefix-operation syntax
 *  that you may have been seen in lisp-like languages.
   ($ IF ($ 0= A1 @ )) ($ THEN ." hello " )
 * Note that an opening simple => ( paren is a comment.
 */
FCode (p4_prefix_begin)
{
  FX (p4_Q_comp);
  FX (p4_tick);
  FX_PUSH (PAREN_MAGIC);
}

/** ) ( cs-token -- )
 * takes the execution-token from => ($ and compiles
 * it using => >COMPILE
 */
FCode (p4_prefix_end)
{
  p4_Q_pairs (PAREN_MAGIC);
  FX (p4_to_compile);
}

/** )) ( cs-token cs-token -- )
 * takes two execution-tokens from two of => ($ and compiles
 * them on after another using => >COMPILE
 simulate:
    : )) [COMPILE] ) [COMPILE] ) ; IMMEDIATE
 */
FCode (p4_prefix_end_doubled)
{
  p4_Q_pairs (PAREN_MAGIC);
  FX (p4_to_compile);
  p4_Q_pairs (PAREN_MAGIC);
  FX (p4_to_compile);
}

/** ((@>)) ( -- value )
 * execution compiled by => @>
 */
FCode (p4_fetch_from_execution)
{
  FX_PUSH ( *P4_POP_(p4cell*,IP));  
}

extern FCode (p4_constant_RT);

/** @> name ( -- value )
 * does fetch the value from the PFA of the named item, which
 * may be about everything, including a => VARIABLE , => VALUE
 * => LVALUE , => LOCALS| , => VAR , => DEFER , => DOER , => DOES>
 * and more.
 */
FCode (p4_fetch_from)
{

  if (!STATE)
  {
     p4xt xt;
     p4_tick (&xt);
     FX_PUSH (*TO_BODY(xt));
  }else{
     char* p; int l;
     char* x; 
      p4_parseword (' ', &p, &l);
      if ((x = p4_find (p, l))) {
         if (*p4_name_from (x) != p4_constant_RT_ ) 
	    COMMA (p4_name_from (x));
         else{
            p4_compile1 ();
            COMMA (TO_BODY (p4_name_from (x)));
         }
         return;
      }
      if (!p4_compile_local (p,l)) {
         p4_hereword (p,l);
         p4_throw (THROW_UNDEFINED);
      }
  }
}
P4COMPILES (p4_fetch_from, p4_fetch_from_execution, P4_SKIPS_CELL, P4_DEFAULT_STYLE);

#ifdef SMART_WORDS

static p4xt
_p4_smart_interpret_char (char c)
{
  auto char wd[] = "interpret- ";
  register char* nfa;
  register p4xt xt;

  wd[sizeof(wd)-2] = c;
  nfa = p4_find (wd, sizeof(wd)-1);
  if (!nfa || !(*_FFA(nfa)|P4xIMMEDIATE)) return 0; /* must be immediate */
  xt = p4_name_from (nfa);
  if (*xt != p4_defer_RT_) return 0; /* must be DEFER or DOER */
  return xt;
}

/**
 */
p4xt 
p4_smart_interpret_char (char c) 
{
  if (!ispunct ((unsigned char) c))
  {
    return NULL;
  }
  else
  {
    return _p4_smart_interpret_char (c);
  }
}

/**
 */
char*
_p4_smart_interpret_init (char c, char* nm, int l)
{
  auto char wd[] = "interpret- ";
  register char* nfa;
  
  wd[sizeof(wd)-2] = c;
  if (!(nfa = p4_find (nm, l))) return 0;
  if (!(*_FFA(nfa)|P4xIMMEDIATE)) return 0;
  
  p4_header_from (p4_defer_RT_, P4xIMMEDIATE, wd, sizeof(wd)-1);
  FX_COMMA (p4_name_from (nfa));
  return nfa;
}

/**
 */
void
p4_smart_interpret_init (char c, char* nm, int l)
{
  if (!_p4_smart_interpret_init (c, nm, l))
     p4_throw (THROW_UNDEFINED);
}

/** SMART-INTERPRET-INIT
 * creates a set of interpret-words that are used in the inner
 * interpreter, so if a word is unknown to the interpreter-loop
 * it will use the first char of that word, attach it to an 
 * "interpret-" prefix, and tries to use that =>'IMMEDIATE'-=>'DEFER'-word 
 * on the rest of the word. This => SMART-INTERPRET-INIT will set up
 * words like interpret-" so you can write 
 * <c>"hello"</c>  instead of   <c>" hello"</c>
 * and it creates interpret-\ so that words like <c>\if-unix</c> are
 * ignoring the line if the word <c>\if-unknown</c> is unknown in itself.
 * This is usually <i>not</i> activated on startup.
 */
FCode (p4_smart_interpret_init)
{
  _p4_smart_interpret_init ('(', "(", 1);   /* so comments will even get shorter */
  _p4_smart_interpret_init ('\\', "\\", 1); /* will silently ignore any unknown backslash word */
  _p4_smart_interpret_init ('@', "@>", 2);
  _p4_smart_interpret_init ('!', "TO", 2);
  _p4_smart_interpret_init ('\'', "ASCII", 5);  /* C like */ 
  _p4_smart_interpret_init ('^', "CONTROL", 7); 
# ifdef P4_C_QUOTE
    _p4_smart_interpret_init ('\"', "C\"", 2); /* C like */ 
# else
    _p4_smart_interpret_init ('\"', "S\"", 2); /* C like */
# endif
}

/** SMART-INTERPRET-OFF
 * disables the SMART-INTERPRET extension in => INTERPRET ,
 * see => SMART-INTERPRET-INIT
 */
FCode (p4_smart_interpret_off)
{
   PFE.smart_char = NULL;
}

/** SMART-INTERPRET-ON
 * enables the SMART-INTERPRET extension in => INTERPRET ,
 * see => SMART-INTERPRET-INIT - the default for smart-interpret 
 * is always off
 */
FCode (p4_smart_interpret_on)
{
   PFE.smart_char = p4_smart_interpret_char;
}

#endif
  /*SMART_WORDS*/
  

#if P4_MP
extern void p4_goto_cpu (int n);
extern void p4_mainloop_cpu (int n);

/** GOTO-CPU ( nr -- )
 * the goto-cpu feature will vanish. It is used to create a 
 * multi-forth in a non-threaded environment. This word
 * used the number and switches the QUIT-interpret loop to
 * use the other cpu's dictionary.
 */
FCode (p4_goto_cpu_RT)
{
  p4_goto_cpu (WP_PFA[0]);
}

/** <<CPU>> ( -- )
 * the goto-cpu feature will vanish. It had been used to create a 
 * multi-forth in a non-threaded environment. This word
 * creates a number of CPU*-immediates that will switch the
 * QUIT-interpret loop to use the other cpu's dictionary.
 */
FCode (p4_load_cpus)
{
  char  name[4] = "CPU0";          /*fixmee MP_MAX bound to 10 */
  char* nfa;
  int i;
  
  for (i=0; i < P4_opt.cpus ; i++)
  {
     name[3] = '0' + i; 

     p4_make_head (name, 4, &nfa, CURRENT)
        ->cfa = p4_goto_cpu_RT_ ;
     COMMA (i);
  }
}

/** CPU# ( -- )
 * the goto-cpu feature will vanish. It is used to create a 
 * multi-forth in a non-threaded environment. This word
 * returns the number of the cpu-dictionary that is 
 * currently in use.
 */
FCode (p4_cpu_nr)
{
  FX_PUSH (PFE.nr);
}
#endif

/* ----------- output convenience ---------- */
extern FCode (p4_emit);

/** 
 *  printing a forth counted string is done through %#s,
 *  the standard %s is ignored defending against invalid use,
 *  but it may be useful to use %1s to print a real
 *  zeroterminated-string.
 */
static int
p4sprintf (char* s)
{
  p4char formbuf[255];
  p4char* formed;
  p4char* format;
  int format_n;
  int argn = 0;
  p4cell argv[16];

  formed = formbuf;
  format = (void*)FX_POP; /* get the argument string */
  format_n = *format++; /* COUNT */
  
  while (format_n > 0)
  {
      if (*format=='%') {
         *formed++ = *format++; format_n--;
          if (*format == '%') {
             *formed++ = *format++; format_n--;
              continue;
          }
          if (*format == 's') {
            /* not sure what a normal string is in this context */
              *formed++ = '%'; /* so it is ignored */
              format++; format_n--;
              FX_DROP;
              continue;
          }
          argv [argn++] = FX_POP; 
          while (format_n > 0)
          {
             if (argn >= 15) goto printnow;
             if (format[0] == '.' && format[1] == '*') {
                argv [argn++] = FX_POP;
                *formed++ = *format++; format_n--;
                *formed++ = *format++; format_n--;
                continue;
             }
             
             if (format[0] == '#' && format[1] == 's') {
                p4char* p = (void*) argv [argn-1];
                argv [argn-1] = *p; argv [argn++] = (p4cell)(p+1); /*COUNT*/
                *formed++ = '.';
                *formed++ = '*';
                *formed++ = 's';
                format+=2; format_n-=2;
                break;
             }
             
             if (isalpha(*format))
                break;
             
             *formed++ = *format++; format_n--;
          }
          continue;
      }
  /*else*/
      *formed++ = *format++; format_n--;
  }
printnow:
  { int printed;
    *formed='\0'; /* should we do it really here ? 
                     or let it do the caller -gud */

    if (argn <= 8)
    {
       printed = sprintf (s, formbuf, 
            argv[0], argv[1], argv[2], argv[3], 
            argv[4], argv[5], argv[6], argv[7]);
    }else{
#    ifdef P4_UPPER_REGS /* i960 */
     _p4_thread_save();
#    endif
     printed = sprintf (s, formbuf, 
        argv[0], argv[1], argv[2], argv[3], 
        argv[4], argv[5], argv[6], argv[7],
        argv[8], argv[9], argv[10], argv[11], 
        argv[12], argv[13], argv[14], argv[15]);

#    ifdef P4_UPPER_REGS
     _p4_thread_load(); 
#    endif
     if (argn > 13)
        { P4_warn1 ("quite many args for a printf (%i)", argn); }
    }
    if (printed > 255)
      { P4_warn1 ("printf long string (%i chars)", printed); }
    return printed;
  }
}

/** SPRINTF ( args ... format$ dest$ -- len-dest ) 
 * just like the standard sprintf() function in C, but
 * the format is a counted string and accepts %#s to
 * be the format-symbol for a forth-counted string.
 * The result is a zeroterminated string at dest$ having
 * a length being returned. To create a forth-counted
 * string, you could use:
   variable A 256 ALLOT
   15 " example" " the %#s value is %i" A 1+ SPRINTF A C!
   A COUNT TYPE
 */
FCode (p4_sprintf)
{
  FX_PUSH (p4sprintf ((void*)FX_POP));
}

/** PRINTF ( args ... format$ -- )
 * uses => SPRINTF to print to a temporary 256-char buffer
 * and prints it to stdout afterwards. See the example
 * at => SPRINTF of what it does internally.
 */
FCode (p4_printf)
{
  char outbuf[256];
  p4sprintf (outbuf);
  p4_outs (outbuf);
}

/** .H2 ( value -- )
 * print hexadecimal, but with per-byte 0-padding
   0x0     -> 00
   0xf     -> 0f
   0x12    -> 12
   0x123   -> 0123
   0x1234  -> 1234
   0x12345 -> 012345
 */
FCode (p4_dot_h2)
{
  p4cell val = FX_POP;
  if (val < 0x100)
     p4_outf ("%02x", val);
  else if (val < 0x10000)
     p4_outf ("%04x", val);
  else if (val < 0x1000000)
     p4_outf ("%06x", val);
  else 
     p4_outf ("%08x", val);
}

/* ---------- */

void
p4_forget_loadf()
{
  char* dp = (char*) WP_PFA[0];

  p4_forget (dp);
}

/** LOADF filename ( -- )
 *  loads a file just like => INCLUDE but does also put
 *  a => MARKER in the => LOADED dictionary that you can
 *  do a => FORGET on to kill everything being loaded
 *  from that file.
 */
FCode (p4_loadf)
{
   extern int p4_included1 (const char* fn, int n, int throws);
   char filename[32];
   char* dp = DP;
   char* fn = p4_word(' ');

   p4_store_c_string (fn+1, *fn, filename, 32);

   if (p4_included1 (fn + 1, *(p4char *)fn, 0))
      p4_forget_word ("%s", (p4cell)filename, p4_forget_loadf, (p4cell)dp);
}

/** @name MAKE-words
 * this make-implementation is quite different from the usual 
 * doer..make implementation. Actually, doer and defer are the
 * same in pfe, ie. make will store the cfa wherever you want it,
 * even in a locals-variable! For that purpose, the make-execution
 * works on the following compiled layout:
   
   +-----------------+----------------+---------------+-----------+
   | (make-exec)-CFA | TO-data-token  | BRANCH-offset | colon-RT  |
   +-----------------+----------------+---------------+-----------+
   
 * note: I had to hack debug.c to work correctly on this. beware.
*/
/**@{*/

/** ((MAKE-))
 * compiled by => MAKE
 */
FCode (p4_make_to_local_execution)
{
  FX_PUSH (IP+2);             /* push following colon-RT, ie. CFA */
  FX (p4_to_local_execution); /* let TO put it into local */
  FX_BRANCH;                  /* and branch over */
}

/** ((MAKE))
 * compiled by => MAKE
 */
FCode (p4_make_to_execution)
{
  FX_PUSH (IP+2);             /* push following colon-RT, ie. CFA */
  FX (p4_to_execution);       /* let TO put it into var/defer */
  FX_BRANCH;                  /* and branch over */
}

/* the 3 in MAKE-begin style is invariant, see debug.c */
#define P4_MAKE0_STYLE 3, 1, 0,  0, 4  /* almost like IF */
#define P4_MAKE1_STYLE 1, 0, -4, 1, 0  /* almost like THEN */

/** MAKE word ... ;AND ( -- )
 * make a seperated piece of code between => MAKE and => ;AND 
 * and on execution of the => MAKE the named word is twisted
 * to point to this piece of code. The word is usually 
 * a => DOER but the current implementation works 
 * on => DEFER just as well, just as it does on other words who
 * expect to find an execution-token in its PFA. You could even
 * create a colon-word that starts with => NOOP and can then make
 * that colon-word be prefixed with the execution of the code piece. 
 * This => MAKE
 * does even work on => LOCALS| and => VAR but it is uncertain
 * what that is good for.
 */
FCode (p4_make)
{
  extern FCode (p4_to);

  if (STATE) 
  {
     FX (p4_to);     /* let TO set up the first two tokens of make-compile */
     FX (p4_forward_mark);  /* third token is empty, filled at ";and"  */
  } else {
     FX_PUSH (DP);	/* value to be */
     FX (p4_tick);	/* stored in ticked cfa */
     *SP += CELLSIZE;   /* ehh, i mean pfa */
  }
  COMMA (p4_colon_RT_); /* the implicit CFA that we need */
  FX_PUSH (STATE); if (!STATE) STATE = P4_TRUE;
  FX_PUSH (P4_MAKE_MAGIC);
}
P4COMPILES2(p4_make, p4_make_to_execution, p4_make_to_local_execution,
        	P4_SKIPS_TO_TOKEN, P4_MAKE0_STYLE);

/** (;AND)
 * compiled by => ;AND
 */
FCode (p4_semicolon_and_execution)
{
   FX (p4_semicolon_execution);
   /* cannot use it in P4COMPILES directly, since it would prevent
      decompiler from acting on at that place 
   */
}

/** ;AND ( -- )
 * For the code piece between => MAKE and => ;AND , this word
 * will do just an => EXIT . For the code outside of
 * the => MAKE construct a branch-around must be resolved then.
 */                
FCode (p4_semicolon_and)
{
  extern FCode (p4_store);

  p4_Q_pairs (P4_MAKE_MAGIC);

  p4_compile1 ();
  STATE = FX_POP;

  if (STATE) 
  {
        FX (p4_forward_resolve);   
  }else{
     	FX (p4_store);
        COMMA (0);
  }
}
P4COMPILES (p4_semicolon_and, p4_semicolon_and_execution,
        	P4_SKIPS_NOTHING, P4_MAKE1_STYLE);
/**@}*/

/** '> name ( -- xt )
 * get the execution-token, ie the CFA, of the word following.
 * This word is fully state-smart while the ANSI standard words
 * namely => ' and => ['] are not.
 */
FCode (p4_tick_from)
{
   extern FCode (p4_bracket_tick);
   extern FCode (p4_tick);
   
   if (STATE)
      FX (p4_bracket_tick);
   else
      FX (p4_tick);
}
P4COMPILES (p4_tick_from, p4_literal_execution, 
   	P4_SKIPS_CELL, P4_DEFAULT_STYLE);
        
/** !> name ( value -- )
 * actually a synonym for => TO but very common amongst
 * forth interpreters
 */
FCode (p4_store_to)
{
  /* copied from to, to help debugger */
  extern FCode (p4_to);
  FX (p4_to);
}
P4COMPILES2 (p4_store_to, p4_to_execution, p4_to_local_execution,
        	P4_SKIPS_TO_TOKEN, P4_DEFAULT_STYLE);

/* these are mostly copied from p4_to in core.c */        
        
/** ((INTO))
 * execution compiled by => INTO
 */
FCode (p4_into_execution)
{
  FX_PUSH (TO_BODY (P4_POP(IP)));
}

/** ((INTO-)) ( -- value ) 
 * execution compiled by => INTO
 */
FCode (p4_into_local_execution)
{
  FX_PUSH (&LP[ P4_POP(IP) ]);
}

/** INTO word ( -- pfa )
 * will return the parameter-field address of the following word.
 * Unlike others, this word will also return the address of
 * => LOCALS| and local => VAR - so in fact a <c>TO A</c> and 
 * <c>INTO A !</c> are the same. This word is most useful when calling
 * C-exported function with a temporary local-VAR as a return-place
 * argument - so the address of a local has to be given as an arg.
 * Beware that you should not try to save the address anywhere else,
 * since a local's address does always depend of the RP-depth.
 */
FCode (p4_into)
{
  char *p;
  int l, n;
  p4xt xt;

  if (STATE)
    {
      p = p4_word (' ');
      l = *(p4char *) p++;
      if (PFE.locals && (n = p4_find_local (p, l)) != 0)
	{
	  p4_compile2 ();
	  COMMA (n);
	}
      else
	{
	  if ((p = p4_find (p, l)) == NULL)
	    p4_throw (THROW_UNDEFINED);
	  p4_compile1 ();
	  COMMA (p4_name_from (p));
	}
    }
  else
    {
      p4_tick (&xt);
      FX_PUSH (TO_BODY (xt));
    }
}

P4COMPILES2 (p4_into, p4_into_execution, p4_into_local_execution,
	   P4_SKIPS_TO_TOKEN, P4_DEFAULT_STYLE);

/** ((0LVALUE))
 * compiled by => VAR
 */           
FCode (p4_zero_local_RT)
{
   LP[ P4_POP(IP) ] = 0;
}

/** VAR name ( -- )
 * a fully state-smart version, that does either create a => VALUE
 * or => LVALUE and does always initialize it with null - so
 * unlike either => VALUE and => LVALUE this word does not take
 * any stack-argument. The name itself stems from javascript.
 */
FCode (p4_var)
{
  extern FCode (p4_value);
  extern FCode (p4_local_value);
  
  if (!STATE) 
  {     	/* same as 0 VALUE in interpreting mode, a global VALUE */
     FX_PUSH (0);
     FX (p4_value);
  }else{	/* same as 0 LVALUE in compiling mode, a LOCAL value */
     FX (p4_local_value);  /* calls compile2 for initializer */
  }
}
P4COMPILES2 (p4_var, p4_to_execution, p4_zero_local_RT,
   	P4_SKIPS_TO_TOKEN, P4_NEW1_STYLE);
        
FCode (p4_offset_RT)
{
  *SP += WP_PFA[0];
}
 
FCode (p4_offset_colon)
{
  p4_header ( PFX(p4_offset_RT), 0);
  COMMA (*SP); FX_DROP;
}

/** [DEFINED] word ( -- nfa|0 ) immediate
 * does check for the word using find (so it does not throw like => ' )
 * and puts it on stack. As it is immediate it does work in compile-mode
 * too, so it places its argument in the cs-stack then. This is most
 * useful with a directly following => [IF] clause, so that sth. like
 * an <c>[IFDEF] word</c> can be simulated through <c>[DEFINED] word [IF]</c>
 */
FCode (p4_bracket_defined)
{
  char* p;
  int   l;
  
  p4_parseword (' ', &p, &l);
  FX_PUSH (p4_find (p, l));
}
extern FCode(p4_defer);

/** [NOT] ( a -- a' )
 * executes => 0= but this word is immediate so that it does 
 * affect the cs-stack while compiling rather than compiling
 * anything. This is useful just before words like => [IF] to
 * provide semantics of an <c>[IFNOT]</c>. It is most useful in
 * conjunction with "=> [DEFINED] word" as it the sequence
 * "<c>[DEFINED] word [NOT] [IF]</c>" can simulate "<c>[IFNOTDEF] word</c>"
 */
FCode (p4_bracket_not)
{
  extern FCode(p4_zero_equal);
  FX (p4_zero_equal);
}

LISTWORDS (your) =
{
  CO ("(IMMEDIATE#)", 	p4_immediate_bit),
  CO ("(SMUDGE#)",    	p4_smudge_bit),
  CO (">COMPILE", 	p4_to_compile),
  CI ("($",  		p4_prefix_begin),
  CI (")",		p4_prefix_end),
  CS ("@>",             p4_fetch_from),
  CS ("!>",             p4_store_to), 
  CS ("'>",		p4_tick_from),
  CS ("INTO",		p4_into),
  CS ("VAR",		p4_var),
#ifdef SMART_WORDS
  CO ("SMART-INTERPRET-INIT", p4_smart_interpret_init),
  CO ("SMART-INTERPRET-OFF",  p4_smart_interpret_off),
  CO ("SMART-INTERPRET-ON",   p4_smart_interpret_on),
#endif
#if P4_MP > 1
  CX ("__cpus__",      p4_load_cpus),
  CO ("CPU#",          p4_cpu_nr),
#endif
  CO (".H2",		p4_dot_h2),
  CO ("PRINTF",         p4_printf),
  CO ("SPRINTF",        p4_sprintf),
  CO ("LOADF",		p4_loadf),
  CO ("DOER",           p4_defer),
  CS ("MAKE",           p4_make),
  CS (";AND",           p4_semicolon_and),
  CI ("[DEFINED]",      p4_bracket_defined),
  CI ("[NOT]",          p4_bracket_not),
  CO ("OFFSET:",        p4_offset_colon),  
};
COUNTWORDS (your, "PFE-SMART Kernel extensions");

/*@}*/

